import csv
import os
import sys
import tempfile
import click
import matplotlib.pyplot as plt
import pandas as pd
from sklearn import preprocessing
from sklearn.metrics import silhouette_score
from sklearn.mixture import GaussianMixture
from nidm.experiment.Query import GetProjectsUUID
from nidm.experiment.tools.click_base import cli
from nidm.experiment.tools.rest import RestParser
from .utils import Reporter


@cli.command()
@click.option(
    "--nidm_file_list",
    "-nl",
    required=True,
    help="A comma separated list of NIDM files with full path",
)
@click.option(
    "--var",
    "-variables",
    required=True,
    help='This parameter is for the variables the user would like to complete the k-means algorithm on.\nThe way this looks in the command is python3 nidm_kmeans.py -nl MTdemog_aseg_v2.ttl -v "fs_003343,age*sex,sex,age,group,age*group,bmi"',
)
@click.option(
    "--k_range",
    "-k",
    required=True,
    help="The maximum number of clusters to try. The algorithm will go from 2 to this number to determine the optimal number of clusters.",
)
@click.option(
    "--optimal_cluster_method",
    "-m",
    required=True,
    help="The criterion used to select the optimal partitioning (either Silhouette Score, AIC, or BIC).",
)
@click.option(
    "--output_file",
    "-o",
    required=False,
    help="Optional output file (TXT) to store results of the linear regression, contrast, and regularization",
)
def gmm(nidm_file_list, output_file, var, k_range, optimal_cluster_method):
    """
    This function provides a tool to complete k-means clustering on NIDM data.
    """
    global v  # Needed to do this because the code only used the parameters in the first method, meaning I had to move it all to method 1.
    v = (
        var.strip()
    )  # used in data_aggregation, kmenas(), spaces stripped from left and right
    with Reporter(output_file) as reporter:
        global n  # used in data_aggregation()
        n = nidm_file_list
        global k_num
        k_num = int(k_range.strip())
        global cm
        cm = optimal_cluster_method
        data_aggregation(reporter)
        dataparsing(reporter)
        cluster_number()


def data_aggregation(reporter):  # all data from all the files is collected
    """This function provides query support for NIDM graphs."""
    # if there is a CDE file list, seed the CDE cache
    if v:  # ex: age,sex,DX_GROUP
        print("*" * 107)
        command = (
            "pynidm k-means -nl "
            + n
            + ' -variables "'
            + v
            + '" '
            + "-k "
            + str(k_num)
            + " -m "
            + cm
        )

        reporter.print("Your command was:", command)
        verbosity = 0
        restParser = RestParser(verbosity_level=int(verbosity))
        restParser.setOutputFormat(RestParser.OBJECT_FORMAT)
        global df_list  # used in dataparsing()
        df_list = []
        # set up uri to do fields query for each nidm file
        global file_list
        file_list = n.split(",")
        df_list_holder = {}
        for i in range(len(file_list)):
            df_list_holder[i] = []
        df_holder = {}
        for i in range(len(file_list)):
            df_holder[i] = []
        global condensed_data_holder
        condensed_data_holder = {}
        for i in range(len(file_list)):
            condensed_data_holder[i] = []

        count = 0
        not_found_count = 0
        for nidm_file in file_list:
            # get project UUID
            project = GetProjectsUUID([nidm_file])
            # split the model into its constituent variables
            global var_list
            # below, we edit the model so it splits by +,~, or =. However, to help it out in catching everything
            # we replaced ~ and = with a + so that we can still use split. Regex wasn't working.
            var_list = [vv.strip() for vv in v.split(",")]
            # set the dependent variable to the one dependent variable in the model
            global variables  # used in dataparsing()
            variables = ""
            for i in range(len(var_list) - 1, -1, -1):
                if (
                    "*" not in var_list[i]
                ):  # removing the star term from the columns we're about to pull from data
                    variables = variables + var_list[i] + ","
                else:
                    print(
                        "Interacting variables are not present in clustering models. They will be removed."
                    )
            variables = variables[0 : len(variables) - 1]
            uri = (
                "/projects/"
                + project[0].toPython().split("/")[-1]
                + "?fields="
                + variables
            )
            # get fields output from each file and concatenate
            df_list_holder[count].append(pd.DataFrame(restParser.run([nidm_file], uri)))
            # global dep_var
            df = pd.concat(df_list_holder[count])
            with tempfile.NamedTemporaryFile(
                delete=False
            ) as temp:  # turns the dataframe into a temporary csv
                df.to_csv(temp.name + ".csv")
                temp.close()

            with open(temp.name + ".csv", encoding="utf-8") as fp:
                data = list(
                    csv.reader(fp)
                )  # makes the csv a 2D list to make it easier to call the contents of certain cells

            var_list = variables.split(",")  # makes a list of the independent variables
            numcols = (len(data) - 1) // (
                len(var_list)
            )  # Finds the number of columns in the original dataframe
            condensed_data_holder[count] = [
                [0] * (len(var_list))
            ]  # makes an array 1 row by the number of necessary columns
            for _ in range(
                numcols
            ):  # makes the 2D array big enough to store all of the necessary values in the edited dataset
                condensed_data_holder[count].append([0] * (len(var_list)))
            var_list = [v.split("/")[-1] for v in var_list]
            for i, vr in enumerate(var_list):
                # stores the independent variable names in the first row
                condensed_data_holder[count][0][i] = vr
            numrows = 1  # begins at the first row to add data
            fieldcolumn = (
                0  # the column the variable name is in in the original dataset
            )
            valuecolumn = 0  # the column the value is in in the original dataset
            datacolumn = 0  # if it is identified by the dataElement name instead of the field's name
            not_found_list = []
            for i in range(len(data[0])):
                if (
                    data[0][i] == "sourceVariable"
                ):  # finds the column where the variable names are
                    fieldcolumn = i
                elif (
                    data[0][i] == "source_variable"
                ):  # finds the column where the variable names are
                    fieldcolumn = i
                elif data[0][i] == "isAbout":
                    aboutcolumn = i
                elif data[0][i] == "label":
                    namecolumn = i  # finds the column where the variable names are
                elif data[0][i] == "value":
                    valuecolumn = i  # finds the column where the values are
                elif (
                    data[0][i] == "dataElement"
                ):  # finds the column where the data element is if necessary
                    datacolumn = i
            for i in range(
                len(condensed_data_holder[count][0])
            ):  # starts iterating through the dataset, looking for the name in that
                for j in range(
                    1, len(data)
                ):  # column, so it can append the values under the proper variables
                    try:
                        if (
                            data[j][fieldcolumn] == condensed_data_holder[count][0][i]
                        ):  # in the dataframe, the name is in column 3
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn
                            ]  # in the dataframe, the value is in column 2
                            numrows = (
                                numrows + 1
                            )  # moves on to the next row to add the proper values
                        elif data[j][aboutcolumn] == condensed_data_holder[count][0][i]:
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn
                            ]  # in the dataframe, the value is in column 2
                            numrows = (
                                numrows + 1
                            )  # moves on to the next row to add the proper values
                        elif (
                            condensed_data_holder[count][0][i] in data[j][aboutcolumn]
                        ):  # this is in case the uri only works by querying the part after the last backslash
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn
                            ]  # in the dataframe, the value is in column 2
                            numrows = (
                                numrows + 1
                            )  # moves on to the next row to add the proper values
                        elif (
                            data[j][namecolumn] == condensed_data_holder[count][0][i]
                        ):  # in the dataframe, the name is in column 12
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn
                            ]  # in the dataframe, the value is in column 2
                            numrows = (
                                numrows + 1
                            )  # moves on to the next row to add the proper values
                        elif (
                            condensed_data_holder[count][0][i] == data[j][datacolumn]
                        ):  # in the dataframe, the name is in column 9
                            condensed_data_holder[count][numrows][i] = data[j][
                                valuecolumn
                            ]  # in the dataframe, the value is in column 2
                            numrows = (
                                numrows + 1
                            )  # moves on to the next row to add the proper values
                    except IndexError:
                        numrows = numrows + 1
                numrows = 1  # resets to the first row for the next variable
            temp_list = condensed_data_holder[count]
            for j in range(
                len(temp_list[0]) - 1, 0, -1
            ):  # if the software appends a column with 0 as the heading, it removes this null column
                if temp_list[0][j] == "0" or temp_list[0][j] == "NaN":
                    for row in condensed_data_holder[count]:
                        row.pop(j)
            rowsize = len(condensed_data_holder[count][0])
            count1 = 0
            for i in range(0, rowsize):
                for row in condensed_data_holder[count]:
                    if row[i] == 0 or row[i] == "NaN" or row[i] == "0":
                        count1 = count1 + 1
                if count1 > len(condensed_data_holder[count]) - 2:
                    not_found_list.append(condensed_data_holder[count][0][i])
                count1 = 0
            for i, cdh in enumerate(condensed_data_holder[count][0]):
                if " " in cdh:
                    condensed_data_holder[count][0][i] = cdh.replace(" ", "_")
            var_list = [vr.split("/")[-1].replace(" ", "_") for vr in var_list]
            count += 1
            if len(not_found_list) > 0:
                print("*" * 107)
                print()
                reporter.print("Your variables were " + v)
                reporter.print()
                reporter.print(
                    "The following variables were not found in "
                    + nidm_file
                    + ". The model cannot run because this will skew the data. Try checking your spelling or use nidm_query.py to see other possible variables."
                )
                for i, nf in enumerate(not_found_list):
                    reporter.print(f"{i + 1}. {nf}")
                not_found_list.clear()
                not_found_count += 1
                print()
        if not_found_count > 0:
            sys.exit(1)

    else:
        print("ERROR: No query parameter provided.  See help:")
        print()
        os.system("pynidm k-means --help")
        sys.exit(1)


def dataparsing(
    reporter,
):  # The data is changed to a format that is usable by the linear regression method
    global condensed_data
    condensed_data = []
    for i in range(0, len(file_list)):
        condensed_data = condensed_data + condensed_data_holder[i]
    global k_num
    if len(condensed_data[0]) <= k_num:
        print(
            "\nThe maximum number of clusters specified is greater than the amount of data present."
        )
        print(
            "The algorithm cannot run with this, so k_num will be reduced to 1 less than the length of the dataset."
        )
        k_num = len(condensed_data) - 1
        print("The k_num value is now: " + str(k_num))
    x = pd.read_csv(
        opencsv(condensed_data)
    )  # changes the dataframe to a csv to make it easier to work with
    x.head()  # prints what the csv looks like
    x.dtypes  # checks data format
    obj_df = x.select_dtypes  # puts all the variables in a dataset
    x.shape  # says number of rows and columns in form of tuple
    x.describe()  # says dataset statistics
    obj_df = x.select_dtypes(
        include=["object"]
    ).copy()  # takes everything that is an object (not float or int) and puts it in a new dataset
    obj_df.head()  # prints the new dataset
    int_df = x.select_dtypes(
        include=["int64"]
    ).copy()  # takes everything that is an int and puts it in a new dataset
    float_df = x.select_dtypes(
        include=["float64"]
    ).copy()  # takes everything that is a float and puts it in a new dataset
    df_int_float = pd.concat([float_df, int_df], axis=1)
    stringvars = []  # starts a list that will store all variables that are not numbers
    for i in range(1, len(condensed_data)):  # goes through each variable
        for j in range(len(condensed_data[0])):  # in the 2D array
            try:  # if the value of the field can be turned into a float (is numerical)
                float(condensed_data[i][j])  # this means it's a number
            except ValueError:  # if it can't be (is a string)
                if (
                    condensed_data[0][j] not in stringvars
                ):  # adds the variable name to the list if it isn't there already
                    stringvars.append(condensed_data[0][j])
    le = (
        preprocessing.LabelEncoder()
    )  # anything involving le shows the encoding of categorical variables
    for sv in stringvars:
        le.fit(obj_df[sv].astype(str))
    obj_df_trf = obj_df.astype(str).apply(
        le.fit_transform
    )  # transforms the categorical variables into numbers.
    global df_final  # also used in linreg()
    if not obj_df_trf.empty:
        df_final = pd.concat(
            [df_int_float, obj_df_trf], axis=1
        )  # join_axes=[df_int_float.index])
    else:
        df_final = df_int_float
    reporter.print(df_final.to_string(header=True, index=True))
    reporter.print("\n\n" + ("*" * 107))
    reporter.print("\n\nModel Results: ")


def cluster_number():
    index = 0
    global levels  # also used in contrasting()
    levels = []
    for i in range(1, len(condensed_data)):
        if condensed_data[i][index] not in levels:
            levels.append(condensed_data[i][index])
    levels = list(range(len(levels)))

    # Beginning of the linear regression
    global X
    # global y
    # Unsure on how to proceed here with interacting variables, since I'm sure dmatrices won't work

    # scaler = MinMaxScaler()
    #
    # for i in range(len(model_list)):
    #     scaler.fit(df_final[[model_list[i]]])
    #     df_final[[model_list[i]]] = scaler.transform(df_final[[model_list[i]]])
    X = df_final[var_list]
    if "si" in cm.lower():
        print("Sillhoute Score")

        ss = []

        for i in range(2, k_num):
            model = GaussianMixture(n_components=i, init_params="kmeans")
            cluster_labels = model.fit_predict(X)
            silhouette_avg = silhouette_score(X, cluster_labels)
            ss.append(silhouette_avg)
        optimal_i = 0
        distance_to_one = abs(1 - ss[0])
        for i, s in enumerate(ss):
            if abs(1 - s) <= distance_to_one:
                optimal_i = i
                distance_to_one = abs(1 - s)

        n_clusters = optimal_i + 2
        print(
            "Optimal number of clusters: " + str(n_clusters)
        )  # optimal number of clusters
        gmm = GaussianMixture(n_components=n_clusters).fit(X)
        labels = gmm.fit(X).predict(X)
        ax = None or plt.gca()
        X = df_final[var_list].to_numpy()
        ax.scatter(X[:, 0], X[:, 1], c=labels, s=40, cmap="viridis", zorder=2)
        ax.axis("equal")
        plt.show()

    if "a" in cm.lower():
        print("AIC\n")
        aic = []
        for i in range(2, k_num):
            model = GaussianMixture(n_components=i, init_params="kmeans")
            model.fit(X)
            aic.append(model.bic(X))
        min_aic = aic[0]
        min_i = 0
        for i in range(1, len(aic)):
            if aic[i] <= min_aic:
                min_aic = aic[i]
                min_i = i
        n_clusters = min_i + 2
        print(
            "Optimal number of clusters: " + str(n_clusters)
        )  # optimal number of clusters, minimizing aic

        # min_aic = aic[0]
        # max_aic = aic[0]
        # max_i = 0
        # min_i = 0
        # for i in range(1, len(aic)):
        #     if aic[i] >= max_aic:
        #         max_aic = aic[i]
        #         max_i = i
        #     elif aic[i] <= min_aic:
        #         min_aic = aic[i]
        #         min_i = i
        # p1 = np.array([min_i, aic[min_i]])
        # p2 = np.array([max_i, aic[max_i]])
        # # the way I am doing the method is as follows:
        # # the different sse values form a curve like an L (like an exponential decay)
        # # The elbow is the point furthest from a line connecting max and min
        # # So I am calculating the distance, and the maximum distance from point to curve shows the optimal point
        # # AKA the number of clusters
        # dist = []
        # for n in range(0, len(aic)):
        #     norm = np.linalg.norm
        #     p3 = np.array([n, aic[n]])
        #     dist.append(np.abs(norm(np.cross(p2 - p1, p1 - p3))) / norm(p2 - p1))
        # max_dist = dist[0]
        # n_clusters = 2
        # for x in range(1, len(dist)):
        #     if dist[x] >= max_dist:
        #         max_dist = dist[x]
        #         n_clusters = x + 2
        #
        # plt.plot(aic)
        # plt.show()

        gmm = GaussianMixture(n_components=n_clusters).fit(X)
        labels = gmm.fit(X).predict(X)
        ax = None or plt.gca()
        X = df_final[var_list].to_numpy()
        ax.scatter(X[:, 0], X[:, 1], c=labels, s=40, cmap="viridis", zorder=2)
        ax.axis("equal")
        plt.show()

    if "b" in cm.lower():
        print("\n\nBIC\n")
        bic = []
        for i in range(2, k_num):
            model = GaussianMixture(n_components=i, init_params="kmeans")
            model.fit(X)
            bic.append(model.bic(X))
        min_bic = bic[0]
        min_i = 0
        for i in range(1, len(bic)):
            if bic[i] <= min_bic:
                min_bic = bic[i]
                min_i = i
        n_clusters = min_i + 2
        # min_bic = bic[0]
        # max_bic = bic[0]
        # max_i = 0
        # min_i = 0
        # for i in range(1,len(bic)):
        #     if bic[i]>=max_bic:
        #         max_bic = bic[i]
        #         max_i = i
        #     elif bic[i]<= min_bic:
        #         min_bic = bic[i]
        #         min_i = i
        # p1 = np.array([min_i, bic[min_i]])
        # p2 = np.array([max_i, bic[max_i]])
        # # the way I am doing the method is as follows:
        # # the different sse values form a curve like an L (like an exponential decay)
        # # The elbow is the point furthest from a line connecting max and min
        # # So I am calculating the distance, and the maximum distance from point to curve shows the optimal point
        # # AKA the number of clusters
        # dist = []
        # for n in range(0, len(bic)):
        #     norm = np.linalg.norm
        #     p3 = np.array([n, bic[n]])
        #     dist.append(np.abs(norm(np.cross(p2 - p1, p1 - p3))) / norm(p2 - p1))
        # max_dist = dist[0]
        # n_clusters = 2
        # for x in range(1, len(dist)):
        #     if dist[x] >= max_dist:
        #         max_dist = dist[x]
        #         n_clusters = x + 2
        # plt.plot(bic)
        # plt.show()
        print("Optimal number of clusters:", n_clusters)
        gmm = GaussianMixture(n_components=n_clusters).fit(X)
        labels = gmm.fit(X).predict(X)
        ax = plt.gca()
        X = df_final[var_list].to_numpy()
        ax.scatter(X[:, 0], X[:, 1], c=labels, s=40, cmap="viridis", zorder=2)
        ax.axis("equal")
        plt.show()


def opencsv(data):
    """saves a list of lists as a csv and opens"""
    handle, fn = tempfile.mkstemp(suffix=".csv")
    with os.fdopen(
        handle, "w", encoding="utf8", errors="surrogateescape", newline=""
    ) as f:
        writer = csv.writer(f)
        writer.writerows(data)
    return fn


# it can be used calling the script `python nidm_query.py -nl ... -q ..
if __name__ == "__main__":
    gmm()



from __future__ import annotations
from dataclasses import dataclass
from typing import Optional
import pytest
from nidm.experiment.tools.rest import RestParser
from nidm.util import urlretrieve
from ..conftest import BrainVol


@dataclass
class OpenNeuro:
    files: list[str]
    project_uri: str


@pytest.fixture(scope="module")
def openneuro(tmp_path_factory: pytest.TempPathFactory) -> OpenNeuro:
    tmp_path = tmp_path_factory.mktemp("openneuro")
    urlretrieve(
        "https://raw.githubusercontent.com/dbkeator/simple2_NIDM_examples/master/datasets.datalad.org/openneuro/ds000120/nidm.ttl",
        tmp_path / "ds000120.nidm.ttl",
    )
    files = [str(tmp_path / "ds000120.nidm.ttl")]
    restParser = RestParser(output_format=RestParser.OBJECT_FORMAT)
    projects2 = restParser.run(files, "/projects")
    project_uri: Optional[str] = None
    for p in projects2:
        proj_info = restParser.run(files, f"/projects/{p}")
        if (
            "dctypes:title" in proj_info.keys()
            and proj_info["dctypes:title"]
            == "Developmental changes in brain function underlying the influence of reward processing on inhibitory control (Slot Reward)"
        ):
            project_uri = p
    assert project_uri is not None
    return OpenNeuro(files, project_uri)


def test_rest_sub_id(brain_vol: BrainVol, openneuro: OpenNeuro) -> None:
    restParser = RestParser()
    restParser.setOutputFormat(RestParser.OBJECT_FORMAT)

    result = restParser.run(
        brain_vol.files + openneuro.files,
        f"/projects/{brain_vol.cmu_test_project_uuid}",
    )

    sub_id = result["subjects"]["subject id"][5]
    sub_uuid = result["subjects"]["uuid"][5]

    result2 = restParser.run(brain_vol.files + openneuro.files, f"/subjects/{sub_id}")

    # make sure we got the same UUID when looking up by sub id
    assert result2["uuid"] == sub_uuid
    assert len(result2["instruments"]) > 0



#!/usr/bin/env python
""" Definition of constants

@author: Camille Maumet <c.m.j.maumet@warwick.ac.uk>
@copyright: University of Warwick 2014
@author: David Keator <dbkeator@uci.edu>
    Added Python provtoolbox  support
    10/3/17 Modified Namespace to be QualifiedName for provtoolbox support...left most of the NIDM-Results Namespaces the same
@author: Sanu Ann Abraham <sanuann@mit.edu>
        05/04/2018 Added python ProvONE support
"""
from collections import namedtuple
from prov.constants import PROV_ATTRIBUTE_LITERALS, PROV_ATTRIBUTE_QNAMES, PROV_N_MAP
from prov.model import Namespace as provNamespace
from prov.model import ProvDocument, QualifiedName
from rdflib import Graph, Namespace

DD = namedtuple("DD", ["source", "variable"])

PROV = Namespace("http://www.w3.org/ns/prov#")
PROVONE = provNamespace(
    "provone", "http://purl.dataone.org/provone/2015/01/15/ontology#"
)

NIDM_URL = "http://purl.org/nidash/nidm#"
NIDM = Namespace(NIDM_URL)

NIIRI = Namespace("http://iri.nidash.org/")
AFNI = Namespace("http://purl.org/nidash/afni#")
SPM = Namespace("http://purl.org/nidash/spm#")
FSL = Namespace("http://purl.org/nidash/fsl#")
FREESURFER = Namespace("https://surfer.nmr.mgh.harvard.edu/")
ANTS = Namespace("http://stnava.github.io/ANTs/")
RDFS = Namespace("http://www.w3.org/2000/01/rdf-schema#")
CRYPTO = Namespace(
    "http://id.loc.gov/vocabulary/preservation/\
cryptographicHashFunctions#"
)
DC = Namespace("http://purl.org/dc/elements/1.1/")
DCT = Namespace("http://purl.org/dc/terms/")
OWL = Namespace("http://www.w3.org/2002/07/owl#")
XSD = Namespace("http://www.w3.org/2001/XMLSchema#")

OBO_URL = "http://purl.obolibrary.org/obo/"
OBO = Namespace(OBO_URL)
# Added by DBK for NIDM-Experiment 1/13/17
NFO = Namespace("http://www.semanticdesktop.org/ontologies/2007/03/22/nfo#")
SCR = Namespace("http://scicrunch.org/resolver/")
NLX = Namespace("http://uri.neuinfo.org/nif/nifstd/")
SKOS = Namespace("http://www.w3.org/2004/02/skos/core#")
FOAF = Namespace("http://xmlns.com/foaf/0.1/")
VC = Namespace("http://www.w3.org/2006/vcard/ns#")
DICOM = Namespace("http://neurolex.org/wiki/Category/DICOM_term/")
DCTYPES = Namespace("http://purl.org/dc/dcmitype/")
NCIT = Namespace("http://ncitt.ncit.nih.gov/")
DCAT = Namespace("http://www.w3.org/ns/dcat#")
BIRNLEX = Namespace("http://bioontology.org/projects/ontologies/birnlex/")
NDAR = Namespace("https://ndar.nih.gov/api/datadictionary/v2/dataelement/")
NCICB = Namespace("http://ncicb.nci.nih.gov/xml/owl/EVS/Thesaurus.owl#")
SIO = Namespace("http://semanticscience.org/ontology/sio.owl#")
BIDS = Namespace("http://bids.neuroimaging.io/")
OWL = Namespace("http://www.w3.org/2002/07/owl#")
ONLI = Namespace("http://neurolog.unice.fr/ontoneurolog/v3.0/instrument.owl#")
PATO = Namespace("http://purl.obolibrary.org/obo/pato#")
DATALAD = Namespace("http://datasets.datalad.org/")
INTERLEX = Namespace("http://uri.interlex.org/")
EDAM = Namespace("https://bioportal.bioontology.org/ontologies/EDAM")

namespaces = {
    # "prov": PROV,
    "nidm": NIDM,
    "niiri": NIIRI,
    "afni": AFNI,
    "spm": SPM,
    "fsl": FSL,
    "freesurfer": FREESURFER,
    "ants": ANTS,
    "rdfs": RDFS,
    "crypto": CRYPTO,
    "dct": DCT,
    "obo": OBO,
    "nfo": NFO,
    "dc": DC,
    "nlx": NLX,
    "scr": SCR,
    "foaf": FOAF,
    "vc": VC,
    "dicom": DICOM,
    "dctypes": DCTYPES,
    "ncit": NCIT,
    "dcat": DCAT,
    "birnlex": BIRNLEX,
    "ndar": NDAR,
    "ncicb": NCICB,
    "sio": SIO,
    "bids": BIDS,
    "owl": OWL,
    "onli": ONLI,
    "pato": PATO,
    "datalad": DATALAD,
    "ilx": INTERLEX,
    "edam": EDAM,
}

# Empty graph used to compute qnames
q_graph = Graph()
for name, namespace in namespaces.items():
    q_graph.bind(name, namespace)


# DBK Added - Empty graph using provtoolbox used to compute qnames
# dj: changed to a new class
class NIDMDocument(ProvDocument):
    def __init__(self, namespaces=None):
        if namespaces is not None:
            super().__init__(namespaces=namespaces)
        else:
            super().__init__()


# NIDM constants
FSL_GAMMAHRF = FSL["FSL_0000007"]
FSL_FSLS_GAMMA_HRF = FSL["FSL_0000006"]
NIDM_HAS_MRI_PROTOCOL = NIDM["NIDM_0000172"]
NIDM_NUMBER_OF_SUBJECTS = NIDM["NIDM_0000171"]
NIDM_GROUP_NAME = NIDM["NIDM_0000170"]
NIDM_DATA = NIDM["NIDM_0000169"]
NIDM_SPM_RESULTS_NIDM = NIDM["NIDM_0000168"]
NIDM_NIDMFSL = NIDM["NIDM_0000167"]
NIDM_NIDM_RESULTS_EXPORT = NIDM["NIDM_0000166"]
NIDM_NIDM_RESULTS_EXPORTER = NIDM["NIDM_0000165"]
NIDM_NEUROIMAGING_ANALYSIS_SOFTWARE = NIDM["NIDM_0000164"]
NIDM_CONTRAST_EXPLAINED_MEAN_SQUARE_MAP = NIDM["NIDM_0000163"]
NIDM_THRESHOLD = NIDM["NIDM_0000162"]
NIDM_EQUIVALENT_THRESHOLD = NIDM["NIDM_0000161"]
NIDM_P_VALUE_UNCORRECTED = NIDM["NIDM_0000160"]
NIDM_P_VALUE_UNCORRECTED_QNAME = q_graph.qname(NIDM_P_VALUE_UNCORRECTED)
NIDM_NOISE_FWHM_IN_VOXELS = NIDM["NIDM_0000159"]
NIDM_NOISE_FWHM_IN_VERTICES = NIDM["NIDM_0000158"]
NIDM_NOISE_FWHM_IN_UNITS = NIDM["NIDM_0000157"]
FSL_FEAT_VERSION = FSL["FSL_0000005"]
FSL_DRIFT_CUTOFF_PERIOD = FSL["FSL_0000004"]
FSL_TEMPORAL_DERIVATIVE = FSL["FSL_0000003"]
FSL_GAUSSIAN_RUNNING_LINE_DRIFT_MODEL = FSL["FSL_0000002"]
FSL_FSLS_GAMMA_DIFFERENCE_HRF = FSL["FSL_0000001"]
SPM_PARTIAL_CONJUNCTION_DEGREE = SPM["SPM_0000015"]
SPM_SMALLEST_SUPRA_THRESHOLD_CLUSTER_SIZE_IN_VOXELS_FWE05 = SPM["SPM_0000014"]
SPM_SMALLEST_SUPRA_THRESHOLD_CLUSTER_SIZE_IN_VOXELS_FDR05 = SPM["SPM_0000013"]
SPM_SMALLEST_SUPRA_THRESHOLD_CLUSTER_SIZE_IN_VERTICES_FWE05 = SPM["SPM_0000012"]
SPM_SMALLEST_SUPRA_THRESHOLD_CLUSTER_SIZE_IN_VERTICES_FDR05 = SPM["SPM_0000011"]
SPM_SEARCH_VOLUME_RESELS_GEOMETRY = SPM["SPM_0000010"]
SPM_TEMPORAL_DERIVATIVE = SPM["SPM_0000006"]
SPM_KCONJUNCTION_INFERENCE = SPM["SPM_0000005"]
SPM_CANONICAL_HRF = SPM["SPM_0000004"]
SPM_DISPERSION_DERIVATIVE = SPM["SPM_0000003"]
SPM_DCT_DRIFT_MODEL = SPM["SPM_0000002"]
SPM_SPMS_DRIFT_CUT_OFF_PERIOD = SPM["SPM_0000001"]
NIDM_CLUSTERSIZEINRESELS = NIDM["NIDM_0000156"]
NIDM_F_MRI_DESIGN = NIDM["NIDM_0000155"]
NIDM_MIXED_DESIGN = NIDM["NIDM_0000154"]
NIDM_EVENT_RELATED_DESIGN = NIDM["NIDM_0000153"]
NIDM_BLOCK_BASED_DESIGN = NIDM["NIDM_0000152"]
NIDM_SINE_BASIS_SET = NIDM["NIDM_0000151"]
NIDM_LINEAR_SPLINE_BASIS_SET = NIDM["NIDM_0000150"]
NIDM_SEARCH_VOLUME_IN_RESELS = NIDM["NIDM_0000149"]
NIDM_RESEL_SIZE_IN_VOXELS = NIDM["NIDM_0000148"]
NIDM_HEIGHT_CRITICAL_THRESHOLD_FWE_05 = NIDM["NIDM_0000147"]
NIDM_HEIGHT_CRITICAL_THRESHOLD_FDR_05 = NIDM["NIDM_0000146"]
NIDM_NOISE_ROUGHNESS_IN_VOXELS = NIDM["NIDM_0000145"]
NIDM_RESELS_PER_VOXEL_MAP = NIDM["NIDM_0000144"]
NIDM_EXPECTED_NUMBER_OF_VOXELS_PER_CLUSTER = NIDM["NIDM_0000143"]
NIDM_EXPECTED_NUMBER_OF_VERTICES_PER_CLUSTER = NIDM["NIDM_0000142"]
NIDM_EXPECTED_NUMBER_OF_CLUSTERS = NIDM["NIDM_0000141"]
NIDM_CLUSTER_CENTER_OF_GRAVITY = NIDM["NIDM_0000140"]
NIDM_COORDINATE_VECTOR_IN_VOXELS = NIDM["NIDM_0000139"]
NIDM_HAS_MAXIMUM_INTENSITY_PROJECTION = NIDM["NIDM_0000138"]
NIDM_SEARCH_VOLUME_IN_VERTICES = NIDM["NIDM_0000137"]
NIDM_SEARCH_VOLUME_IN_UNITS = NIDM["NIDM_0000136"]
NIDM_CONTRAST_VARIANCE_MAP = NIDM["NIDM_0000135"]
NIDM_WITH_ESTIMATION_METHOD = NIDM["NIDM_0000134"]
NIDM_VOXEL_UNITS = NIDM["NIDM_0000133"]
NIDM_VOXEL_TO_WORLD_MAPPING = NIDM["NIDM_0000132"]
NIDM_VOXEL_SIZE = NIDM["NIDM_0000131"]
NIDM_VOXEL6CONNECTED = NIDM["NIDM_0000130"]
NIDM_VOXEL26CONNECTED = NIDM["NIDM_0000129"]
NIDM_VOXEL18CONNECTED = NIDM["NIDM_0000128"]
NIDM_VERSION = NIDM["NIDM_0000127"]
NIDM_VARIANCE_SPATIAL_MODEL = NIDM["NIDM_0000126"]
NIDM_USER_SPECIFIED_THRESHOLD_TYPE = NIDM["NIDM_0000125"]
NIDM_TARGET_INTENSITY = NIDM["NIDM_0000124"]
NIDM_STATISTIC_TYPE = NIDM["NIDM_0000123"]
NIDM_SOFTWARE_VERSION = NIDM["NIDM_0000122"]
NIDM_SEARCH_VOLUME_IN_VOXELS = NIDM["NIDM_0000121"]
NIDM_RANDOM_FIELD_STATIONARITY = NIDM["NIDM_0000120"]
NIDM_Q_VALUE_FDR = NIDM["NIDM_0000119"]
NIDM_PIXEL8CONNECTED = NIDM["NIDM_0000118"]
NIDM_PIXEL4CONNECTED = NIDM["NIDM_0000117"]
NIDM_P_VALUE_UNCORRECTED = NIDM["NIDM_0000116"]
NIDM_P_VALUE_FWER = NIDM["NIDM_0000115"]
NIDM_P_VALUE = NIDM["NIDM_0000114"]
NIDM_OBJECT_MODEL = NIDM["NIDM_0000113"]
NIDM_NUMBER_OF_DIMENSIONS = NIDM["NIDM_0000112"]
NIDM_NUMBER_OF_CLUSTERS = NIDM["NIDM_0000111"]
NIDM_GAUSSIAN_HRF = NIDM["NIDM_0000110"]
NIDM_MIN_DISTANCE_BETWEEN_PEAKS = NIDM["NIDM_0000109"]
NIDM_MAX_NUMBER_OF_PEAKS_PER_CLUSTER = NIDM["NIDM_0000108"]
NIDM_MASKED_MEDIAN = NIDM["NIDM_0000107"]
NIDM_IS_USER_DEFINED = NIDM["NIDM_0000106"]
NIDM_IN_WORLD_COORDINATE_SYSTEM = NIDM["NIDM_0000105"]
NIDM_IN_COORDINATE_SPACE = NIDM["NIDM_0000104"]
NIDM_HAS_MAP_HEADER = NIDM["NIDM_0000103"]
NIDM_HAS_HRF_BASIS = NIDM["NIDM_0000102"]
NIDM_HAS_ERROR_DISTRIBUTION = NIDM["NIDM_0000101"]
NIDM_HAS_ERROR_DEPENDENCE = NIDM["NIDM_0000100"]
NIDM_HAS_CONNECTIVITY_CRITERION = NIDM["NIDM_0000099"]
NIDM_HAS_CLUSTER_LABELS_MAP = NIDM["NIDM_0000098"]
NIDM_HAS_ALTERNATIVE_HYPOTHESIS = NIDM["NIDM_0000097"]
NIDM_GRAND_MEAN_SCALING = NIDM["NIDM_0000096"]
NIDM_ERROR_VARIANCE_HOMOGENEOUS = NIDM["NIDM_0000094"]
NIDM_ERROR_DEGREES_OF_FREEDOM = NIDM["NIDM_0000093"]
NIDM_EQUIVALENT_ZSTATISTIC = NIDM["NIDM_0000092"]
NIDM_EFFECT_DEGREES_OF_FREEDOM = NIDM["NIDM_0000091"]
NIDM_DIMENSIONS_IN_VOXELS = NIDM["NIDM_0000090"]
NIDM_DEPENDENCE_SPATIAL_MODEL = NIDM["NIDM_0000089"]
NIDM_HAS_DRIFT_MODEL = NIDM["NIDM_0000088"]
NIDM_DRIFT_MODEL = NIDM["NIDM_0000087"]
NIDM_COORDINATE_VECTOR = NIDM["NIDM_0000086"]
NIDM_CONTRAST_NAME = NIDM["NIDM_0000085"]
NIDM_CLUSTER_SIZE_IN_VOXELS = NIDM["NIDM_0000084"]
NIDM_CLUSTER_SIZE_IN_VERTICES = NIDM["NIDM_0000083"]
NIDM_CLUSTER_LABEL_ID = NIDM["NIDM_0000082"]
NIDM_WORLD_COORDINATE_SYSTEM = NIDM["NIDM_0000081"]
NIDM_VOXEL_CONNECTIVITY_CRITERION = NIDM["NIDM_0000080"]
NIDM_TWO_TAILED_TEST = NIDM["NIDM_0000079"]
NIDM_TALAIRACH_COORDINATE_SYSTEM = NIDM["NIDM_0000078"]
NIDM_SUBJECT_COORDINATE_SYSTEM = NIDM["NIDM_0000077"]
NIDM_STATISTIC_MAP = NIDM["NIDM_0000076"]
NIDM_STANDARDIZED_COORDINATE_SYSTEM = NIDM["NIDM_0000075"]
NIDM_SPATIALLY_REGULARIZED_MODEL = NIDM["NIDM_0000074"]
NIDM_SPATIALLY_LOCAL_MODEL = NIDM["NIDM_0000073"]
NIDM_SPATIALLY_GLOBAL_MODEL = NIDM["NIDM_0000072"]
NIDM_SPATIAL_MODEL = NIDM["NIDM_0000071"]
NIDM_SUPRA_THRESHOLD_CLUSTER = NIDM["NIDM_0000070"]
NIDM_FOURIER_BASIS_SET = NIDM["NIDM_0000069"]
NIDM_SEARCH_SPACE_MASK_MAP = NIDM["NIDM_0000068"]
NIDM_CUSTOM_BASIS_SET = NIDM["NIDM_0000067"]
NIDM_RESIDUAL_MEAN_SQUARES_MAP = NIDM["NIDM_0000066"]
NIDM_POISSON_DISTRIBUTION = NIDM["NIDM_0000065"]
NIDM_PIXEL_CONNECTIVITY_CRITERION = NIDM["NIDM_0000064"]
NIDM_PEAK_DEFINITION_CRITERIA = NIDM["NIDM_0000063"]
NIDM_PEAK = NIDM["NIDM_0000062"]
NIDM_PARAMETER_ESTIMATE_MAP = NIDM["NIDM_0000061"]
NIDM_ONE_TAILED_TEST = NIDM["NIDM_0000060"]
NIDM_NON_PARAMETRIC_SYMMETRIC_DISTRIBUTION = NIDM["NIDM_0000059"]
NIDM_NON_PARAMETRIC_DISTRIBUTION = NIDM["NIDM_0000058"]
NIDM_NIDM_OBJECT_MODEL = NIDM["NIDM_0000057"]
NIDM_MODEL_PARAMETERS_ESTIMATION = NIDM["NIDM_0000056"]
NIDM_MNI305_COORDINATE_SYSTEM = NIDM["NIDM_0000055"]
NIDM_MASK_MAP = NIDM["NIDM_0000054"]
NIDM_MAP_HEADER = NIDM["NIDM_0000053"]
NIDM_MAP = NIDM["NIDM_0000052"]
NIDM_MNI_COORDINATE_SYSTEM = NIDM["NIDM_0000051"]
NIDM_IXI549_COORDINATE_SYSTEM = NIDM["NIDM_0000050"]
NIDM_INFERENCE = NIDM["NIDM_0000049"]
NIDM_INDEPENDENT_ERROR = NIDM["NIDM_0000048"]
NIDM_ICBM_MNI152_NON_LINEAR6TH_GENERATION_COORDINATE_SYSTEM = NIDM["NIDM_0000047"]
NIDM_ICBM_MNI152_NON_LINEAR2009C_SYMMETRIC_COORDINATE_SYSTEM = NIDM["NIDM_0000046"]
NIDM_ICBM_MNI152_NON_LINEAR2009C_ASYMMETRIC_COORDINATE_SYSTEM = NIDM["NIDM_0000045"]
NIDM_ICBM_MNI152_NON_LINEAR2009B_SYMMETRIC_COORDINATE_SYSTEM = NIDM["NIDM_0000044"]
NIDM_ICBM_MNI152_NON_LINEAR2009B_ASYMMETRIC_COORDINATE_SYSTEM = NIDM["NIDM_0000043"]
NIDM_ICBM_MNI152_NON_LINEAR2009A_SYMMETRIC_COORDINATE_SYSTEM = NIDM["NIDM_0000042"]
NIDM_ICBM_MNI152_NON_LINEAR2009A_ASYMMETRIC_COORDINATE_SYSTEM = NIDM["NIDM_0000041"]
NIDM_ICBM_MNI152_LINEAR_COORDINATE_SYSTEM = NIDM["NIDM_0000040"]
NIDM_ICBM452_WARP5_COORDINATE_SYSTEM = NIDM["NIDM_0000039"]
NIDM_ICBM452_AIR_COORDINATE_SYSTEM = NIDM["NIDM_0000038"]
NIDM_HEMODYNAMIC_RESPONSE_FUNCTION_DERIVATIVE = NIDM["NIDM_0000037"]
NIDM_HEMODYNAMIC_RESPONSE_FUNCTION_BASIS = NIDM["NIDM_0000036"]
NIDM_HEMODYNAMIC_RESPONSE_FUNCTION = NIDM["NIDM_0000035"]
NIDM_HEIGHT_THRESHOLD = NIDM["NIDM_0000034"]
NIDM_GRAND_MEAN_MAP = NIDM["NIDM_0000033"]
NIDM_GAMMA_HRF = NIDM["NIDM_0000031"]
NIDM_GAMMA_HRB = NIDM["NIDM_0000030"]
NIDM_GAMMA_DIFFERENCE_HRF = NIDM["NIDM_0000029"]
NIDM_FINITE_IMPULSE_RESPONSE_HRB = NIDM["NIDM_0000028"]
NIDM_RESULTS = NIDM["NIDM_0000027"]
NIDM_EXTENT_THRESHOLD = NIDM["NIDM_0000026"]
NIDM_EXCURSION_SET_MAP = NIDM["NIDM_0000025"]
NIDM_EXCHANGEABLE_ERROR = NIDM["NIDM_0000024"]
NIDM_ERROR_MODEL = NIDM["NIDM_0000023"]
NIDM_ERROR_DISTRIBUTION = NIDM["NIDM_0000022"]
NIDM_REGRESSOR_NAMES = NIDM["NIDM_0000021"]
NIDM_DISPLAY_MASK_MAP = NIDM["NIDM_0000020"]
NIDM_DESIGN_MATRIX = NIDM["NIDM_0000019"]
NIDM_CUSTOM_COORDINATE_SYSTEM = NIDM["NIDM_0000017"]
NIDM_COORDINATE_SPACE = NIDM["NIDM_0000016"]
NIDM_COORDINATE = NIDM["NIDM_0000015"]
NIDM_LEGENDRE_POLYNOMIAL_ORDER = NIDM["NIDM_0000014"]
NIDM_CONTRAST_STANDARD_ERROR_MAP = NIDM["NIDM_0000013"]
NIDM_CONNECTIVITY_CRITERION = NIDM["NIDM_0000012"]
NIDM_CONJUNCTION_INFERENCE = NIDM["NIDM_0000011"]
NIDM_HAS_FMRI_DESIGN = NIDM["NIDM_0000010"]
NIDM_COLIN27_COORDINATE_SYSTEM = NIDM["NIDM_0000009"]
NIDM_CLUSTER_LABELS_MAP = NIDM["NIDM_0000008"]
NIDM_CLUSTER_DEFINITION_CRITERIA = NIDM["NIDM_0000007"]
NIDM_CLUSTER = NIDM["NIDM_0000006"]
NIDM_BINOMIAL_DISTRIBUTION = NIDM["NIDM_0000005"]
NIDM_BINARY_MAP = NIDM["NIDM_0000004"]
NIDM_CONTRAST_ESTIMATION = NIDM["NIDM_0000001"]
NIDM_CONTRAST_MAP = NIDM["NIDM_0000002"]


# NIDM-Experiment##############################################################

NIDM_DATAELEMENT = QualifiedName(provNamespace("nidm", NIDM), "DataElement")
NIDM_PROJECT = QualifiedName(provNamespace("nidm", NIDM), "Project")
# NIDM_PROJECT_TYPE = QualifiedName(provNamespace("dctypes", DCTYPES),"Dataset")
NIDM_PROJECT_IDENTIFIER = QualifiedName(provNamespace("sio", SIO), "Identifier")
NIDM_PROJECT_NAME = QualifiedName(provNamespace("dctypes", DCTYPES), "title")
NIDM_PROJECT_DESCRIPTION = QualifiedName(provNamespace("dct", DCT), "description")
NIDM_DESCRIPTION = QualifiedName(provNamespace("dct", DCT), "description")
NIDM_DEFINITION = QualifiedName(provNamespace("dct", DCT), "description")
NIDM_PROJECT_LICENSE = QualifiedName(provNamespace("dct", DCT), "license")
NIDM_PROJECT_URL = QualifiedName(provNamespace("sio", SIO), "URL")
NIDM_PROJECT_REFERENCES = QualifiedName(provNamespace("dcat", DCAT), "creator")
NIDM_AUTHOR = QualifiedName(provNamespace("ncit", DCAT), "author")
NIDM_SESSION = QualifiedName(provNamespace("nidm", NIDM), "Session")
NIDM_ACQUISITION_ACTIVITY = QualifiedName(provNamespace("nidm", NIDM), "Acquisition")
NIDM_ACQUISITION_MODALITY = QualifiedName(
    provNamespace("nidm", NIDM), "hadAcquisitionModality"
)
NIDM_ASSESSMENT_ACQUISITION = QualifiedName(
    provNamespace("onli", ONLI), "instrument-based-assessment"
)
NIDM_ACQUISITION_ENTITY = QualifiedName(
    provNamespace("nidm", NIDM), "AcquisitionObject"
)

NIDM_PROJECT_SOURCE = QualifiedName(provNamespace("dctypes", DCTYPES), "source")
NIDM_HAD_NUMERICAL_VALUE = QualifiedName(
    provNamespace("nidm", NIDM), "hadNumericalValue"
)
NIDM_BATH_SOLUTION = QualifiedName(provNamespace("nidm", NIDM), "BathSolution")
NIDM_CELL_TYPE = QualifiedName(provNamespace("nidm", NIDM), "CellType")
NIDM_CHANNEL_NUMBER = QualifiedName(provNamespace("nidm", NIDM), "ChannelNumber")
NIDM_ELECTRODE_IMPEDANCE = QualifiedName(
    provNamespace("nidm", NIDM), "ElectrodeImpedance"
)
NIDM_GROUP_LABEL = QualifiedName(provNamespace("nidm", NIDM), "GroupLabel")
NIDM_HOLLOW_ELECTRODE_SOLUTION = QualifiedName(
    provNamespace("nidm", NIDM), "HollowElectrodeSolution"
)
NIDM_HAD_IMAGE_CONTRACT_TYPE = QualifiedName(
    provNamespace("nidm", NIDM), "hadImageContractType"
)
NIDM_HAD_IMAGE_USAGE_TYPE = QualifiedName(
    provNamespace("nidm", NIDM), "hadImageUsageType"
)
NIDM_NUBMER_OF_CHANNELS = QualifiedName(provNamespace("nidm", NIDM), "NubmerOfChannels")
NIDM_APPLIED_FILTER = QualifiedName(provNamespace("nidm", NIDM), "AppliedFilter")
NIDM_SOLUTION_FLOW_SPEED = QualifiedName(
    provNamespace("nidm", NIDM), "SolutionFlowSpeed"
)
NIDM_RECORDING_LOCATION = QualifiedName(
    provNamespace("nidm", NIDM), "RecordingLocation"
)

NIDM_DEMOGRAPHICS_ENTITY = QualifiedName(
    provNamespace("nidm", NIDM), "DemographicsInstrument"
)
NIDM_ASSESSMENT_USAGE_TYPE = QualifiedName(
    provNamespace("nidm", NIDM), "AssessmentUsageType"
)


NIDM_ASSESSMENT_ENTITY = QualifiedName(
    provNamespace("onli", ONLI), "assessment-instrument"
)
# files
NIDM_FILENAME = QualifiedName(provNamespace("nfo", NFO), "filename")
NIDM_FILE = QualifiedName(provNamespace("sio", SIO), "file")
# Roles
NIDM_PI = QualifiedName(provNamespace("birnlex", BIRNLEX), "birnlex_2152")
NIDM_COI = QualifiedName(provNamespace("birnlex", BIRNLEX), "birnlex_2199")
NIDM_PARTICIPANT = QualifiedName(provNamespace("sio", SIO), "Subject")
# Demographics
NIDM_AGE = QualifiedName(provNamespace("ncidb", NCICB), "Age")
NIDM_GENDER = QualifiedName(provNamespace("ndar", NDAR), "gender")
NIDM_SEX = QualifiedName(provNamespace("pato", PATO), "PhenotypicSex")
NIDM_HANDEDNESS = QualifiedName(provNamespace("obo", OBO), "handedness")
# NIDM_HANDEDNESS = OBO["PATO_0002201"] is correct term ID for handedness above
NIDM_ETHNICITY = QualifiedName(provNamespace("sio", SIO), "ethnicity")
NIDM_RACE = QualifiedName(provNamespace("sio", SIO), "race")

# NCICB_ETHNICITY = NCICB["C16564"] is correct term ID for ethnic group
NIDM_DIAGNOSIS = QualifiedName(provNamespace("ncit", NCIT), "Diagnosis")
NIDM_FAMILY_NAME = QualifiedName(provNamespace("foaf", FOAF), "familyName")
NIDM_GIVEN_NAME = QualifiedName(provNamespace("foaf", FOAF), "givenName")
NIDM_SUBJECTID = QualifiedName(provNamespace("ndar", NDAR), "src_subject_id")
# MRI scan types
NIDM_IMAGE_CONTRAST_TYPE = QualifiedName(
    provNamespace("nidm", NIDM), "hadImageContrastType"
)
NIDM_IMAGE_USAGE_TYPE = QualifiedName(provNamespace("nidm", NIDM), "hadImageUsageType")
NIDM_PET = QualifiedName(provNamespace("nidm", NIDM), "PositronEmissionTomography")
NIDM_MRI = QualifiedName(provNamespace("nidm", NIDM), "MagneticResonanceImaging")
NIDM_MRI_ANATOMIC_SCAN = QualifiedName(provNamespace("nidm", NIDM), "Anatomical")
NIDM_MRI_STRUCTURE_SCAN = QualifiedName(provNamespace("nidm", NIDM), "Structural")
NIDM_MRI_FUNCTION_SCAN = QualifiedName(provNamespace("nidm", NIDM), "Functional")
NIDM_MRI_DWI_SCAN = QualifiedName(provNamespace("nidm", NIDM), "DiffusionWeighted")
NIDM_MRI_DWI_BVAL = QualifiedName(provNamespace("nidm", NIDM), "b-value")
NIDM_MRI_DWI_BVEC = QualifiedName(provNamespace("nidm", NIDM), "b-vector")
NIDM_MRI_FUNCTION_TASK = QualifiedName(provNamespace("nidm", NIDM), "Task")
NIDM_MRI_T1 = QualifiedName(provNamespace("nidm", NIDM), "T1Weighted")
NIDM_MRI_T2 = QualifiedName(provNamespace("nidm", NIDM), "T2Weighted")
NIDM_MRI_T2_STAR = QualifiedName(provNamespace("nidm", NIDM), "T2StarWeighted")
NIDM_MRI_DIFFUSION_TENSOR = QualifiedName(
    provNamespace("nidm", NIDM), "DiffusionTensor"
)
NIDM_MRI_FLOW = QualifiedName(provNamespace("nidm", NIDM), "FlowWeighted")
NIDM_MRI_BOLD_EVENTS = QualifiedName(
    provNamespace("nidm", NIDM), "StimulusResponseFile"
)
NIDM_MRI_ASL = QualifiedName(provNamespace("nidm", NIDM), "ArterialSpinLabeling")
CRYPTO_SHA512 = QualifiedName(provNamespace("crypto", CRYPTO), "sha512")
DATALAD_LOCATION = QualifiedName(provNamespace("datalad", DATALAD), "Location")
NIDM_DOI = QualifiedName(provNamespace("edam", EDAM), "data_1188")
NIDM_FUNDING = QualifiedName(provNamespace("obo", OBO), "IAO_0000623")
NIDM_ACKNOWLEDGEMENTS = QualifiedName(provNamespace("obo", OBO), "IAO_0000324")
##############################################################################
# OBO constants
OBO_EXAMPLE = OBO["IAO_0000112"]
OBO_TERM_EDITOR = OBO["IAO_0000117"]
OBO_EDITOR_NOTE = OBO["IAO_0000116"]

OBO_PENDING_FINAL = OBO["IAO_0000125"]
OBO_METADATA_COMPLETE = OBO["IAO_0000120"]
OBO_METADATA_INCOMPLETE = OBO["IAO_0000123"]
OBO_REQUIRES_DISCUSSION = OBO["IAO_0000428"]
OBO_UNCURATED = OBO["IAO_0000124"]
OBO_TO_BE_REPLACED = OBO["IAO_0000423"]
OBO_READY = OBO["IAO_0000122"]
OBO_DEFINITION = OBO["IAO_0000115"]

OBO_STATISTIC = OBO["STATO_0000039"]
OBO_STATISTIC_QNAME = q_graph.qname(OBO_STATISTIC)
OBO_P_VALUE_FWER = OBO["OBI_0001265"]
OBO_P_VALUE_FWER_QNAME = q_graph.qname(OBO_P_VALUE_FWER)
OBO_Q_VALUE_FDR = OBO["OBI_0001442"]
OBO_Q_VALUE_FDR_QNAME = q_graph.qname(OBO_Q_VALUE_FDR)

HAS_CURATION_STATUS = OBO["IAO_0000114"]

STATO_OLS = OBO["STATO_0000370"]
STATO_OLS_STR = q_graph.qname(STATO_OLS)
# TODO: labels should be grabbed automatically from the corresponding owl file
STATO_OLS_LABEL = "obo:'ordinary least squares estimation'"
STATO_GLS = OBO["STATO_0000372"]
STATO_GLS_STR = q_graph.qname(STATO_GLS)
STATO_GLS_LABEL = "obo:'generalized least squares estimation'"
STATO_TSTATISTIC = OBO["STATO_0000176"]
STATO_TSTATISTIC_STR = q_graph.qname(STATO_TSTATISTIC)
STATO_TSTATISTIC_LABEL = "obo:'t-statistic'"
STATO_ZSTATISTIC = OBO["STATO_0000376"]
STATO_ZSTATISTIC_STR = q_graph.qname(STATO_ZSTATISTIC)
STATO_ZSTATISTIC_LABEL = "obo:'Z-statistic'"
STATO_CONTRAST_WEIGHT_MATRIX = OBO["STATO_0000323"]
STATO_GAUSSIAN_DISTRIBUTION = OBO["STATO_0000227"]
STATO_UNSTRUCTURED_COVARIANCE = OBO["STATO_0000405"]
STATO_GROUP = OBO["STATO_0000193"]

SPM_SOFTWARE = SCR["SCR_007037"]
FSL_SOFTWARE = SCR["SCR_002823"]

NLX_MRI_SCANNER = NLX["birnlex_2100"]
NLX_FMRI_PROTOCOL = NLX["birnlex_2250"]
NLX_IMAGING_INSTRUMENT = NLX["birnlex_2094"]

SKOS_DEFINITION = SKOS["definition"]


# ProvONE Constants for classes
PROVONE_PROCESS = PROVONE["Process"]
PROVONE_USER = PROVONE["User"]
PROVONE_PROCESSEXEC = PROVONE["ProcessExec"]
PROVONE_DATA = PROVONE["Data"]
PROVONE_INPUTPORT = PROVONE["InputPort"]
PROVONE_OUTPUTPORT = PROVONE["OutputPort"]
PROVONE_DATALINK = PROVONE["DataLink"]
PROVONE_SEQCTRLLINK = PROVONE["seqCtrlLink"]

# ProvONE Constants for Associations
PROVONE_HASOUTPORT = PROVONE["hasOutPort"]
PROVONE_HASINPORT = PROVONE["hasInPort"]
PROVONE_HASSUBPROCESS = PROVONE["hasSubProcess"]
PROVONE_INPORTTODL = PROVONE["inPortToDL"]
PROVONE_DLTOINPORT = PROVONE["DLToInPort"]
PROVONE_OUTPORTTODL = PROVONE["outPortToDL"]
PROVONE_DLTOOUTPORT = PROVONE["DLToOutPort"]
PROVONE_CLTODESTP = PROVONE["CLtoDestP"]
PROVONE_SOURCEPTOCL = PROVONE["sourcePToCL"]
PROVONE_DATAONLINK = PROVONE["dataOnLink"]
PROVONE_HASDEFAULTPARAM = PROVONE["hasDefaultParameter"]
PROVONE_ISPARTOF = PROVONE["isPartOf"]
PROVONE_MEMBERSHIP = PROVONE["hadMember"]

# ProvONE notation mapping
PROVONE_N_MAP = {
    PROVONE_PROCESS: "process",
    PROVONE_PROCESSEXEC: "processExec",
    PROVONE_USER: "user",
    PROVONE_DATA: "data",
    PROVONE_HASINPORT: "hasInPort",
    PROVONE_INPUTPORT: "inputPort",
    PROVONE_OUTPUTPORT: "outputPort",
    PROVONE_HASOUTPORT: "hasOutPort",
    PROVONE_HASSUBPROCESS: "hasSubProcess",
    PROVONE_INPORTTODL: "inPortToDL",
    PROVONE_DATALINK: "dataLink",
    PROVONE_SEQCTRLLINK: "seqCtrlLink",
    PROVONE_CLTODESTP: "CLtoDestP",
    PROVONE_SOURCEPTOCL: "sourcePtoCL",
    PROVONE_OUTPORTTODL: "outPortToDL",
    PROVONE_DLTOOUTPORT: "DLToOutPort",
    PROVONE_DLTOINPORT: "DLToInPort",
    PROVONE_DATAONLINK: "dataOnLink",
    PROVONE_HASDEFAULTPARAM: "hasDefaultParamter",
    PROVONE_ISPARTOF: "isPartOf",
    PROVONE_MEMBERSHIP: "hadMember",
}

# Identifiers for PROVONE's attributes
PROVONE_ATTR_PROCESS = PROVONE["process"]
PROVONE_ATTR_USER = PROVONE["user"]
PROVONE_ATTR_PROCESSEXEC = PROVONE["processExec"]
PROVONE_ATTR_PLAN = PROVONE["plan"]
PROVONE_ATTR_GENERATED_DATA = PROVONE["generatedData"]
PROVONE_ATTR_USED_DATA = PROVONE["usedData"]
PROVONE_ATTR_GENERATION = PROVONE["generation"]
# PROVONE_ATTR_USAGE = PROVONE['usage']
PROVONE_ATTR_DATA = PROVONE["data"]
PROVONE_ATTR_INFORMED = PROVONE["informed"]
PROVONE_ATTR_INFORMANT = PROVONE["informant"]
PROVONE_ATTR_HASINPORT = PROVONE["hasInPort"]
PROVONE_ATTR_HASOUTPORT = PROVONE["HasOutPort"]
PROVONE_ATTR_INPUTPORT = PROVONE["InputPort"]
PROVONE_ATTR_OUTPUTPORT = PROVONE["OutputPort"]
PROVONE_ATTR_GENERATED_PROCESS = PROVONE["generatedProcess"]
PROVONE_ATTR_USED_PROCESS = PROVONE["usedProcess"]
PROVONE_ATTR_HASSUBPROCESS = PROVONE["hasSubProcess"]
PROVONE_ATTR_DATALINK = PROVONE["dataLink"]
PROVONE_ATTR_SEQCTRLLINK = PROVONE["seqCtrlLink"]
PROVONE_ATTR_CLTODESTP = PROVONE["clToDestP"]
PROVONE_ATTR_SOURCEPTOCL = PROVONE["sourcePtoCL"]
PROVONE_ATTR_RELATED_PREXEC = (PROVONE["relatedProcessExec"],)
PROVONE_ATTR_USED_PREXEC = PROVONE["usedProcessExec"]
PROVONE_ATTR_CHILD_PREXEC = PROVONE["childProcessExec"]


PROVONE_ATTRIBUTE_QNAMES = {
    PROVONE_ATTR_PROCESS,
    PROVONE_ATTR_USER,
    PROVONE_ATTR_PROCESSEXEC,
    PROVONE_ATTR_PLAN,
    PROVONE_ATTR_GENERATED_DATA,
    PROVONE_ATTR_USED_DATA,
    PROVONE_ATTR_DATA,
    PROVONE_ATTR_INFORMED,
    PROVONE_ATTR_INFORMANT,
    PROVONE_ATTR_HASINPORT,
    PROVONE_ATTR_HASOUTPORT,
    PROVONE_ATTR_INPUTPORT,
    PROVONE_ATTR_OUTPUTPORT,
    PROVONE_ATTR_GENERATED_PROCESS,
    PROVONE_ATTR_USED_PROCESS,
    PROVONE_ATTR_HASSUBPROCESS,
    PROVONE_ATTR_DATALINK,
    PROVONE_ATTR_SEQCTRLLINK,
    PROVONE_ATTR_CLTODESTP,
    PROVONE_ATTR_SOURCEPTOCL,
    PROVONE_ATTR_RELATED_PREXEC,
    PROVONE_ATTR_USED_PREXEC,
    PROVONE_ATTR_CHILD_PREXEC,
    # PROV_ATTR_COLLECTION
}


# Set of formal attributes of PROV records
PROVONE_ATTRIBUTES = (
    PROVONE_ATTRIBUTE_QNAMES | PROV_ATTRIBUTE_QNAMES | PROV_ATTRIBUTE_LITERALS
)
PROVONE_RECORD_ATTRIBUTES = [(attr, str(attr)) for attr in PROVONE_ATTRIBUTES]

PROV_RECORD_IDS_MAP = {value: rec_type_id for rec_type_id, value in PROV_N_MAP.items()}
PROVONE_ID_ATTRIBUTES_MAP = dict(PROVONE_RECORD_ATTRIBUTES)
PROVONE_ATTRIBUTES_ID_MAP = {
    attribute: prov_id for (prov_id, attribute) in PROVONE_RECORD_ATTRIBUTES
}


# ADDED BY DBK to make searching NIDM-Experiment Terms easier...temporary, should be done in the OWL file #
nidm_experiment_terms = [
    NIDM_PROJECT,
    NIDM_PROJECT_IDENTIFIER,
    NIDM_PROJECT_NAME,
    NIDM_PROJECT_DESCRIPTION,
    NIDM_PROJECT_LICENSE,
    NIDM_PROJECT_URL,
    NIDM_PROJECT_REFERENCES,
    NIDM_AUTHOR,
    NIDM_SESSION,
    NIDM_ACQUISITION_ACTIVITY,
    NIDM_ACQUISITION_MODALITY,
    NIDM_ASSESSMENT_ACQUISITION,
    NIDM_ACQUISITION_ENTITY,
    NIDM_DEMOGRAPHICS_ENTITY,
    NIDM_ASSESSMENT_ENTITY,
    NIDM_FILENAME,
    NIDM_FILE,
    NIDM_PI,
    NIDM_COI,
    NIDM_PARTICIPANT,
    NIDM_AGE,
    NIDM_GENDER,
    NIDM_SEX,
    NIDM_HANDEDNESS,
    NIDM_RACE,
    NIDM_ETHNICITY,
    NIDM_DIAGNOSIS,
    NIDM_FAMILY_NAME,
    NIDM_GIVEN_NAME,
    NIDM_SUBJECTID,
    NIDM_IMAGE_CONTRAST_TYPE,
    NIDM_IMAGE_USAGE_TYPE,
    NIDM_MRI,
    NIDM_MRI_ANATOMIC_SCAN,
    NIDM_MRI_STRUCTURE_SCAN,
    NIDM_MRI_FUNCTION_SCAN,
    NIDM_MRI_DWI_SCAN,
    NIDM_MRI_DWI_BVAL,
    NIDM_MRI_DWI_BVEC,
    NIDM_MRI_FUNCTION_TASK,
    NIDM_MRI_T1,
    NIDM_MRI_T2,
    NIDM_MRI_T2_STAR,
    NIDM_MRI_DIFFUSION_TENSOR,
    NIDM_MRI_FLOW,
    NIDM_MRI_BOLD_EVENTS,
    NIDM_DOI,
]

# Common isAbout URIs
NIDM_IS_ABOUT_AGE = str(INTERLEX["ilx_0100400"])
NIDM_IS_ABOUT_HANDEDNESS = str(OBO["PATO_0002201"])
NIDM_IS_ABOUT_GENDER = str(INTERLEX["ilx_0101292"])

# REST API constants
NIDM_REST_NUM_SUBJECTS = "number_of_subjects"
NIDM_REST_MAX_AGE = "age_max"
NIDM_REST_MIN_AGE = "age_min"
NIDM_REST_GENDER = "gender"
NIDM_REST_AGE = "age"


# canonical CDE file locations
CDE_FILE_LOCATIONS = [
    "https://raw.githubusercontent.com/ReproNim/fsl_seg_to_nidm/master/fsl_seg_to_nidm/mapping_data/fsl_cde.ttl",
    "https://raw.githubusercontent.com/ReproNim/ants_seg_to_nidm/master/ants_seg_to_nidm/mapping_data/ants_cde.ttl",
    "https://raw.githubusercontent.com/ReproNim/segstats_jsonld/master/segstats_jsonld/mapping_data/fs_cde.ttl",
]



from io import StringIO
import json
from pathlib import Path
import prov
import rdflib
from nidm.core import Constants
from nidm.experiment import Project, Session


def test_1(tmp_path: Path) -> None:
    project = Project()
    # save a turtle file
    with open(tmp_path / "test.ttl", "w", encoding="utf-8") as f:
        f.write(project.serializeTurtle())


def test_2(tmp_path: Path) -> None:
    kwargs = {
        Constants.NIDM_PROJECT_NAME: "FBIRN_PhaseII",
        Constants.NIDM_PROJECT_IDENTIFIER: 9610,
        Constants.NIDM_PROJECT_DESCRIPTION: "Test investigation",
    }
    project = Project(attributes=kwargs)

    with open(tmp_path / "test.ttl", "w", encoding="utf-8") as f:
        f.write(project.serializeTurtle())


def test_sessions_1() -> None:
    project = Project()
    assert not project.sessions

    session1 = Session(project)
    project.add_sessions(session1)
    assert session1.label == project.sessions[0].label

    session2 = Session(project)
    project.add_sessions(session2)
    assert len(project.sessions) == 2
    assert session2.label == project.sessions[1].label


def test_sessions_2() -> None:
    project = Project()
    assert not project.sessions

    session1 = Session(project)
    assert project.sessions[0].label == session1.label


def test_sessions_3() -> None:
    project1 = Project()
    project2 = Project()

    session1 = Session(project1)
    session2 = Session(project2)

    project1.add_sessions(session1)
    project1.add_sessions(session2)

    assert len(project1.sessions) == 2
    assert session2.label == project1.sessions[1].label
    assert session1.label == project1.sessions[0].label


def test_project_noparameters():
    # creating project without parameters
    proj = Project()

    # checking if we created ProvDocument
    assert isinstance(proj.bundle, Constants.NIDMDocument)
    assert issubclass(type(proj.bundle), prov.model.ProvDocument)

    # checking graph namespace
    const_l = list(Constants.namespaces)
    namesp = [i.prefix for i in proj.graph.namespaces]
    assert sorted(const_l) == sorted(namesp)

    # checking type
    proj_type = proj.get_type()
    assert eval(proj_type.provn_representation()) == "prov:Activity"

    # checking length of graph records; it doesn work if all tests are run
    assert len(proj.graph.get_records()) == 1


def test_project_emptygraph():
    # creating project without parameters
    proj = Project(empty_graph=True)

    # checking if we created ProvDocument
    assert isinstance(proj.bundle, Constants.NIDMDocument)

    # checking graph namespace
    namesp = [i.prefix for i in proj.graph.namespaces]
    assert namesp == ["nidm"]

    # checking type
    proj_type = proj.get_type()
    assert eval(proj_type.provn_representation()) == "prov:Activity"

    assert len(proj.graph.get_records()) == 1


def test_project_uuid():
    # creating project without parameters
    proj = Project(uuid="my_uuid")

    # checking if we created ProvDocument
    assert isinstance(proj.bundle, Constants.NIDMDocument)
    assert issubclass(type(proj.bundle), prov.model.ProvDocument)

    # checking graph namespace
    const_l = list(Constants.namespaces)
    namesp = [i.prefix for i in proj.graph.namespaces]
    assert sorted(const_l) == sorted(namesp)

    # checking type
    proj_type = proj.get_type()
    assert eval(proj_type.provn_representation()) == "prov:Activity"

    # checking if uuid is correct
    assert proj.identifier.localpart == "my_uuid"

    # checking length of graph records; it doesn work if all tests are run
    assert len(proj.graph.get_records()) == 1


def test_project_att():
    # creating project without parameters
    proj = Project(
        attributes={prov.model.QualifiedName(Constants.NIDM, "title"): "MyPRoject"}
    )

    # checking if we created ProvDocument
    assert isinstance(proj.bundle, Constants.NIDMDocument)
    assert issubclass(type(proj.bundle), prov.model.ProvDocument)

    # checking graph namespace
    const_l = list(Constants.namespaces)
    namesp = [i.prefix for i in proj.graph.namespaces]
    assert sorted(
        const_l + [rdflib.term.URIRef("http://purl.org/nidash/nidm#prefix")]
    ) == sorted(namesp)

    # checking type
    proj_type = proj.get_type()
    assert eval(proj_type.provn_representation()) == "prov:Activity"

    # checking length of graph records; it doesn work if all tests are run
    assert len(proj.graph.get_records()) == 1


def test_session_noparameters():
    # creating project without parameters and a session to the project
    proj = Project()
    Session(proj)

    # checking if we created ProvDocument
    assert isinstance(proj.bundle, Constants.NIDMDocument)
    assert issubclass(type(proj.bundle), prov.model.ProvDocument)

    # checking if one session is added
    assert len(proj.sessions) != 0

    # checking graph namespace
    const_l = list(Constants.namespaces)
    namesp = [i.prefix for i in proj.graph.namespaces]
    assert sorted(const_l) == sorted(namesp)

    # checking type
    proj_type = proj.get_type()
    assert eval(proj_type.provn_representation()) == "prov:Activity"

    # checking length of graph records; it doesn work if all tests are run
    assert len(proj.graph.get_records()) == 2


def test_jsonld_exports(tmp_path: Path) -> None:
    kwargs = {
        Constants.NIDM_PROJECT_NAME: "FBIRN_PhaseII",
        Constants.NIDM_PROJECT_IDENTIFIER: 9610,
        Constants.NIDM_PROJECT_DESCRIPTION: "Test investigation",
    }
    project = Project(uuid="_123456", attributes=kwargs)

    # save a turtle file
    with open(tmp_path / "test.json", "w", encoding="utf-8") as f:
        f.write(project.serializeJSONLD())

    # load in JSON file
    with open(tmp_path / "test.json", encoding="utf-8") as json_file:
        data = json.load(json_file)

    assert data["Identifier"]["@value"] == "9610"
    # WIP  Read back in json-ld file and check that we have the project info


def test_project_trig_serialization():
    outfile = StringIO()

    kwargs = {
        Constants.NIDM_PROJECT_NAME: "FBIRN_PhaseII",
        Constants.NIDM_PROJECT_IDENTIFIER: 9610,
        Constants.NIDM_PROJECT_DESCRIPTION: "Test investigation",
    }
    project = Project(uuid="_123456", attributes=kwargs)

    # save as trig file with graph identifier Constants.NIDM_Project
    test = project.serializeTrig(identifier=Constants.NIIRI["_996"])
    if not isinstance(test, str):
        outfile.write(test.decode("ASCII"))
    else:
        outfile.write(test)
    outfile.seek(0)

    # WIP: RDFLib doesn't seem to have a Trig parser?!?
    # load back into rdf graph and do assertions
    # project2 = Graph()
    # project2.parse(source=outfile)

    # test some assertion on read file
    # print(project2.serialize(format='turtle').decode('ASCII'))
    # print(project2.serialize(format='trig').decode('ASCII'))


# TODO: checking
# attributes{pm.QualifiedName(Namespace("uci", "https.../"), "mascot"): "bleble", ...}
# (has to be "/" at the end (or #)



""" Tools for working with NIDM-Experiment files """

import os.path
import click
from nidm.experiment.Utils import read_nidm
from nidm.experiment.tools.click_base import cli


# adding click argument parsing
@cli.command()
@click.option(
    "--nidm_file_list",
    "-nl",
    required=True,
    help="A comma separated list of NIDM files with full path",
)
def visualize(nidm_file_list):
    """
    This command will produce a visualization(pdf) of the supplied NIDM files named the same as the input files and
    stored in the same directories.
    """

    for nidm_file in nidm_file_list.split(","):
        # read in nidm file
        project = read_nidm(nidm_file)

        # split path and filename for output file writing
        file_parts = os.path.split(nidm_file)

        # write graph as nidm filename + .pdf
        project.save_DotGraph(
            filename=os.path.join(
                file_parts[0], os.path.splitext(file_parts[1])[0] + ".pdf"
            ),
            format="pdf",
        )


if __name__ == "__main__":
    visualize()



from nidm import __version__
from nidm.experiment.tools.click_base import cli


# adding click argument parsing
@cli.command()
def version():
    """
    This function will print the version of pynidm.
    """
    print(f"PyNIDM Version: {__version__}")


if __name__ == "__main__":
    version()



"""Python implementation of the ProvONE Model Specification

References:

ProvONE: http://vcvcomputing.com/provone/provone.html
"""

import io
import logging
import os
import shutil
import tempfile
from urllib.parse import urlparse
from prov.constants import (
    PROV_ASSOCIATION,
    PROV_ATTR_COLLECTION,
    PROV_ATTR_ENDTIME,
    PROV_ATTR_STARTTIME,
    PROV_ATTR_TIME,
    PROV_ATTR_USAGE,
    PROV_ATTRIBUTION,
    PROV_COMMUNICATION,
    PROV_DERIVATION,
    PROV_GENERATION,
    PROV_MEMBERSHIP,
    PROV_N_MAP,
    PROV_USAGE,
)
from prov.model import (
    PROV_REC_CLS,
    ProvActivity,
    ProvAgent,
    ProvAssociation,
    ProvAttribution,
    ProvCommunication,
    ProvDerivation,
    ProvDocument,
    ProvEntity,
    ProvGeneration,
    ProvMembership,
    ProvRelation,
    ProvUsage,
    _ensure_datetime,
)
import nidm.core.serializers
from .Constants import (
    PROVONE_ATTR_CHILD_PREXEC,
    PROVONE_ATTR_DATA,
    PROVONE_ATTR_DATALINK,
    PROVONE_ATTR_GENERATED_DATA,
    PROVONE_ATTR_GENERATED_PROCESS,
    PROVONE_ATTR_GENERATION,
    PROVONE_ATTR_INFORMANT,
    PROVONE_ATTR_INFORMED,
    PROVONE_ATTR_INPUTPORT,
    PROVONE_ATTR_OUTPUTPORT,
    PROVONE_ATTR_PLAN,
    PROVONE_ATTR_PROCESS,
    PROVONE_ATTR_PROCESSEXEC,
    PROVONE_ATTR_SEQCTRLLINK,
    PROVONE_ATTR_USED_DATA,
    PROVONE_ATTR_USED_PREXEC,
    PROVONE_ATTR_USED_PROCESS,
    PROVONE_ATTR_USER,
    PROVONE_CLTODESTP,
    PROVONE_DATA,
    PROVONE_DATALINK,
    PROVONE_DATAONLINK,
    PROVONE_DLTOINPORT,
    PROVONE_DLTOOUTPORT,
    PROVONE_HASDEFAULTPARAM,
    PROVONE_HASINPORT,
    PROVONE_HASOUTPORT,
    PROVONE_HASSUBPROCESS,
    PROVONE_INPORTTODL,
    PROVONE_INPUTPORT,
    PROVONE_ISPARTOF,
    PROVONE_N_MAP,
    PROVONE_OUTPORTTODL,
    PROVONE_OUTPUTPORT,
    PROVONE_PROCESS,
    PROVONE_PROCESSEXEC,
    PROVONE_SEQCTRLLINK,
    PROVONE_SOURCEPTOCL,
    PROVONE_USER,
)

__author__ = "Sanu Ann Abraham"
__email__ = "sanuann@mit.edu"

logger = logging.getLogger(__name__)

# add ProvOne Notation mapping to Prov_N_MAP dict
# PROV_N_MAP.update(PROVONE_N_MAP)
PROVONE_N_MAP.update(PROV_N_MAP)


class ProvPlan(ProvEntity):
    """
    ProvONE Plan element
    """


class Process(ProvEntity):
    """
    ProvONE Process element"""

    _prov_type = PROVONE_PROCESS


class InputPort(ProvEntity):
    """ProvONE Input Port element"""

    _prov_type = PROVONE_INPUTPORT


class OutputPort(ProvEntity):
    """ProvONE Output Port element"""

    _prov_type = PROVONE_OUTPUTPORT


class Data(ProvEntity):
    """
    basic unit of information consumed or produced by a Process. Multiple Data items may be grouped into a Collection.
    """

    _prov_type = PROVONE_DATA


class DataLink(ProvEntity):
    """ProvONE DataLink Element"""

    _prov_type = PROVONE_DATALINK


class SeqCtrlLink(ProvEntity):
    """ProvONE SeqCtrlLink Element"""

    _prov_type = PROVONE_SEQCTRLLINK


class User(ProvAgent):
    """ProvONE User element."""

    _prov_type = PROVONE_USER


class ProcessExec(ProvActivity):
    """ProvONE Process Execution element."""

    _prov_type = PROVONE_PROCESSEXEC


class Attribution(ProvAttribution):
    """ProvONE Attribution relationship."""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_PROCESS, PROVONE_ATTR_USER)

    _prov_type = PROV_ATTRIBUTION


class Association(ProvAssociation):
    """Provenance Association relationship."""

    FORMAL_ATTRIBUTES = (
        PROVONE_ATTR_PROCESSEXEC,
        PROVONE_ATTR_PROCESS,
        PROVONE_ATTR_PLAN,
    )

    _prov_type = PROV_ASSOCIATION


class Communication(ProvCommunication):
    """Provenance Communication relationship."""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_INFORMED, PROVONE_ATTR_INFORMANT)

    _prov_type = PROV_COMMUNICATION


class Derivation(ProvDerivation):
    """Provenance Derivation relationship."""

    FORMAL_ATTRIBUTES = (
        PROVONE_ATTR_GENERATED_DATA,
        PROVONE_ATTR_USED_DATA,
        PROVONE_ATTR_PROCESSEXEC,
        PROVONE_ATTR_GENERATION,
        PROV_ATTR_USAGE,
    )

    _prov_type = PROV_DERIVATION


class Generation(ProvGeneration):
    """Provenance Generation relationship."""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_DATA, PROVONE_ATTR_PROCESSEXEC, PROV_ATTR_TIME)

    # _prov_type = PROV_GENERATION


class Usage(ProvUsage):
    """Provenance Usage relationship."""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_PROCESSEXEC, PROVONE_ATTR_DATA, PROV_ATTR_TIME)

    # _prov_type = PROV_USAGE


class Partnership(ProvRelation):
    """Provenance Membership relationship."""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_USED_PREXEC, PROVONE_ATTR_CHILD_PREXEC)

    _prov_type = PROVONE_ISPARTOF


# Component 6: Collections
class Membership(ProvMembership):
    """Provenance Membership relationship."""

    FORMAL_ATTRIBUTES = (PROV_ATTR_COLLECTION, PROVONE_ATTR_DATA)

    # _prov_type = PROV_MEMBERSHIP


class HasInput(ProvRelation):
    """ProvONE HasInput Port relationship."""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_PROCESS, PROVONE_ATTR_INPUTPORT)

    _prov_type = PROVONE_HASINPORT


class HasOutput(ProvRelation):
    """ProvONE HasOutput Port relationship."""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_PROCESS, PROVONE_ATTR_OUTPUTPORT)

    _prov_type = PROVONE_HASOUTPORT


class HasSubProcess(ProvRelation):
    """ProvONE Has SubProcess relationship."""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_USED_PROCESS, PROVONE_ATTR_GENERATED_PROCESS)

    _prov_type = PROVONE_HASSUBPROCESS


class InToDL(ProvRelation):
    """ProvONE InPort to DL relationship"""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_INPUTPORT, PROVONE_ATTR_DATALINK)

    _prov_type = PROVONE_INPORTTODL


class OutToDL(ProvRelation):
    """ProvONE Output port to DL relationship"""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_OUTPUTPORT, PROVONE_ATTR_DATALINK)

    _prov_type = PROVONE_OUTPORTTODL


class DLtoOutPort(ProvRelation):
    """ProvONE DL to Output port relationship"""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_DATALINK, PROVONE_ATTR_OUTPUTPORT)

    _prov_type = PROVONE_DLTOOUTPORT


class DLtoInPort(ProvRelation):
    """ProvONE DL to Input port relationship"""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_DATALINK, PROVONE_ATTR_INPUTPORT)

    _prov_type = PROVONE_DLTOINPORT


class CLtoDestP(ProvRelation):
    """ProvONE CLtoDestP relationship."""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_SEQCTRLLINK, PROVONE_ATTR_PROCESS)

    _prov_type = PROVONE_CLTODESTP


class SourcePtoCL(ProvRelation):
    """ProvONE SourcePtoCL relationship."""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_PROCESS, PROVONE_ATTR_SEQCTRLLINK)

    _prov_type = PROVONE_SOURCEPTOCL


class DataLinkage(ProvRelation):
    """ProvONE dataOnLink relationship"""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_DATA, PROVONE_ATTR_DATALINK, PROVONE_ATTR_PROCESS)

    _prov_type = PROVONE_DATAONLINK


class Parameterization(ProvRelation):
    """ProvONE hasDefaultParam relationship."""

    FORMAL_ATTRIBUTES = (PROVONE_ATTR_INPUTPORT, PROVONE_ATTR_DATA)

    _prov_type = PROVONE_HASDEFAULTPARAM


class Workflow(
    Process,
):
    pass


#  Class mappings from PROVONE record type
PROV_REC_CLS.update(
    {
        PROVONE_PROCESS: Process,
        PROVONE_PROCESSEXEC: ProcessExec,
        PROVONE_DATA: Data,
        PROV_ATTRIBUTION: Attribution,
        PROV_ASSOCIATION: Association,
        PROV_COMMUNICATION: Communication,
        PROV_DERIVATION: Derivation,
        PROV_GENERATION: Generation,
        PROV_USAGE: Usage,
        PROVONE_INPUTPORT: InputPort,
        PROVONE_HASINPORT: HasInput,
        PROVONE_OUTPUTPORT: OutputPort,
        PROVONE_HASOUTPORT: HasOutput,
        PROVONE_HASSUBPROCESS: HasSubProcess,
        PROVONE_DATALINK: DataLink,
        PROVONE_INPORTTODL: InToDL,
        PROVONE_SEQCTRLLINK: SeqCtrlLink,
        PROVONE_CLTODESTP: CLtoDestP,
        PROVONE_SOURCEPTOCL: SourcePtoCL,
        PROVONE_OUTPORTTODL: OutToDL,
        PROVONE_DLTOOUTPORT: DLtoOutPort,
        PROVONE_DLTOINPORT: DLtoInPort,
        PROVONE_DATAONLINK: DataLinkage,
        PROVONE_HASDEFAULTPARAM: Parameterization,
        PROVONE_USER: User,
        PROVONE_ISPARTOF: Partnership,
        PROV_MEMBERSHIP: Membership,
    }
)


class ProvONEDocument(ProvDocument):
    """ProvONE Document"""

    def __repr__(self):
        return "<ProvONEDocument>"

    def process(self, identifier, other_attributes=None):
        """
        Creates a new process.

        :param identifier: Identifier for new process.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(PROVONE_PROCESS, identifier, None, other_attributes)

    def user(self, identifier, other_attributes=None):
        """
        Creates a new user.

        :param identifier: Identifier for new user.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(PROVONE_USER, identifier, None, other_attributes)

    def data(self, identifier, other_attributes=None):
        """
        Creates a new data.

        :param identifier: Identifier for new data.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(PROVONE_DATA, identifier, None, other_attributes)

    def attribution(self, process_spec, user, identifier=None, other_attributes=None):
        """
        Creates a new attribution record between a process specification and an user.

        :param process_spec: ProcessSpecification or a string identifier for the process spec (relationship
                source).
        :param user: User or string identifier of the user involved in the
                attribution (relationship destination).
        :param identifier: Identifier for new attribution record.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(
            PROV_ATTRIBUTION,
            identifier,
            {PROVONE_ATTR_PROCESS: process_spec, PROVONE_ATTR_USER: user},
            other_attributes,
        )

    def processExec(
        self, identifier, startTime=None, endTime=None, other_attributes=None
    ):
        """
        Creates a new process execution.

        :param identifier: Identifier for new process execution.
        :param startTime: Optional start time for the process execution (default:
        None).
                Either a :py:class:`datetime.datetime` object or a string that can be
                parsed by :py:func:`dateutil.parser`.
        :param endTime: Optional end time for the process execution (default: None).
                Either a :py:class:`datetime.datetime` object or a string that can be
                parsed by :py:func:`dateutil.parser`.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(
            PROVONE_PROCESSEXEC,
            identifier,
            {
                PROV_ATTR_STARTTIME: _ensure_datetime(startTime),
                PROV_ATTR_ENDTIME: _ensure_datetime(endTime),
            },
            other_attributes,
        )

    def association(
        self,
        process_exec,
        process_spec=None,
        plan=None,
        identifier=None,
        other_attributes=None,
    ):
        """
        Creates a new association record for a process execution.

        :param process_exec: Process Execution or a string identifier for the
                process execution.
        :param process_spec: Process Spec or string identifier of the process
                involved in the association (default: None).
        :param plan: Optionally extra entity to state qualified association through
                an internal plan (default: None).
        :param identifier: Identifier for new association record.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(
            PROV_ASSOCIATION,
            identifier,
            {
                PROVONE_ATTR_PROCESSEXEC: process_exec,
                PROVONE_ATTR_PROCESS: process_spec,
                PROVONE_ATTR_PLAN: plan,
            },
            other_attributes,
        )

    def derivation(
        self,
        generatedData,
        usedData,
        process_exec=None,
        generation=None,
        usage=None,
        identifier=None,
        other_attributes=None,
    ):
        """
        Creates a new derivation record for a generated data from a used data.

        :param generatedData: Data or a string identifier for the generated
                data (relationship source).
        :param usedData: Data or a string identifier for the used data
                (relationship destination).
        :param process_exec: Process execution or string identifier of the
                processExec involved in the derivation (default: None).
        :param generation: Optionally extra activity to state qualified generation
                through a generation (default: None).
        :param usage: XXX (default: None).
        :param identifier: Identifier for new derivation record.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).
        """
        attributes = {
            PROVONE_ATTR_GENERATED_DATA: generatedData,
            PROVONE_ATTR_USED_DATA: usedData,
            PROVONE_ATTR_PROCESSEXEC: process_exec,
            PROVONE_ATTR_GENERATION: generation,
            PROV_ATTR_USAGE: usage,
        }
        return self.new_record(
            PROV_DERIVATION, identifier, attributes, other_attributes
        )

    def generation(
        self, data, process_exec=None, time=None, identifier=None, other_attributes=None
    ):
        """
        Creates a new generation record for a data.

        :param data: Data or a string identifier for the data.
        :param process_exec: Process execution or string identifier of the
        process_exec involved in the generation (default: None).
        :param time: Optional time for the generation (default: None).
                Either a :py:class:`datetime.datetime` object or a string that can be
                parsed by :py:func:`dateutil.parser`.
        :param identifier: Identifier for new generation record.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(
            PROV_GENERATION,
            identifier,
            {
                PROVONE_ATTR_DATA: data,
                PROVONE_ATTR_PROCESSEXEC: process_exec,
                PROV_ATTR_TIME: _ensure_datetime(time),
            },
            other_attributes,
        )

    def usage(
        self, process_exec, data=None, time=None, identifier=None, other_attributes=None
    ):
        """
        Creates a new usage record for a process execution.

        :param process_exec: Process Execution or a string identifier for the
                processExec.
        :param data: Data or string identifier of the data involved in
                the usage relationship (default: None).
        :param time: Optional time for the usage (default: None).
                Either a :py:class:`datetime.datetime` object or a string that can be
                parsed by :py:func:`dateutil.parser`.
        :param identifier: Identifier for new usage record.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(
            PROV_USAGE,
            identifier,
            {
                PROVONE_ATTR_PROCESSEXEC: process_exec,
                PROVONE_ATTR_DATA: data,
                PROV_ATTR_TIME: _ensure_datetime(time),
            },
            other_attributes,
        )

    def communication(
        self, informed, informant, identifier=None, other_attributes=None
    ):
        """
        Creates a new communication record for a process exec.

        :param informed: The informed processExec (relationship destination).
        :param informant: The informing processExec (relationship source).
        :param identifier: Identifier for new communication record.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(
            PROV_COMMUNICATION,
            identifier,
            {PROVONE_ATTR_INFORMED: informed, PROVONE_ATTR_INFORMANT: informant},
            other_attributes,
        )

    def input_port(self, identifier, other_attributes=None):
        """
        Creates a new input port.

        :param identifier: Identifier for new input port.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(PROVONE_INPUTPORT, identifier, None, other_attributes)

    def has_in_ports(self, process, in_ports, identifier=None, other_attributes=None):
        """
        Creates a new input port record for a process.

        :param process: Process or a string identifier for the
                process(relationship source).
        :param in_ports: Input Port or string identifier for the used input port (
                relationship destination).
        :param identifier: Identifier for new input port membership.
        :param other_attributes: Optional other attributes as a dictionary or
                list of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(
            PROVONE_HASINPORT,
            identifier,
            {
                PROVONE_ATTR_PROCESS: process,
                PROVONE_ATTR_INPUTPORT: in_ports,
            },
            other_attributes,
        )

    def output_port(self, identifier, other_attributes=None):
        """
        Creates a new output port.

        :param identifier: Identifier for new output port.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(PROVONE_OUTPUTPORT, identifier, None, other_attributes)

    def has_out_ports(self, process, out_ports, identifier=None, other_attributes=None):
        """
        Creates a new input port record for a process.

        :param process: Process or a string identifier for the
                process(relationship source).
        :param out_ports: Output Port or string identifier for the used output
        port (relationship destination).
        :param identifier: Identifier for new output port membership.
        :param other_attributes: Optional other attributes as a dictionary or
                list of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(
            PROVONE_HASOUTPORT,
            identifier,
            {
                PROVONE_ATTR_PROCESS: process,
                PROVONE_ATTR_OUTPUTPORT: out_ports,
            },
            other_attributes,
        )

    def has_sub_process(
        self, used_process, generated_process, identifier=None, other_attributes=None
    ):
        """
        Creates a new has-sub-process record for a generated process from a
        used process.

        :param used_process: Process or a string identifier for the
        used process (relationship source).
        :param generated_process: Process or a string identifier for the
        generated process (relationship destination).
        :param identifier: Identifier for new sub-process record.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).
        """
        attributes = {
            PROVONE_ATTR_USED_PROCESS: used_process,
            PROVONE_ATTR_GENERATED_PROCESS: generated_process,
        }
        return self.new_record(
            PROVONE_HASSUBPROCESS, identifier, attributes, other_attributes
        )

    def dataLink(self, identifier, other_attributes=None):
        """
        Creates a new data link.

        :param identifier: Identifier for new data link.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(PROVONE_DATALINK, identifier, None, other_attributes)

    def inPortToDL(self, in_port, dt_link, identifier=None, other_attributes=None):
        """

        :param in_port: Input port or a string identifier for the
                in_port(relationship source).
        :param dt_link: Data Link or string identifier for the used data link (
                relationship destination).
        :param identifier: Identifier for new data link membership.
        :param other_attributes: Optional other attributes as a dictionary or
                list of tuples to be added to the record optionally (default: None).

        """
        return self.new_record(
            PROVONE_INPORTTODL,
            identifier,
            {
                PROVONE_ATTR_INPUTPORT: in_port,
                PROVONE_ATTR_DATALINK: dt_link,
            },
            other_attributes,
        )

    def outPortToDL(self, out_port, dt_link, identifier=None, other_attributes=None):
        """

        :param out_port: Output port or a string identifier for the
                out_port(relationship source).
        :param dt_link: Data Link or string identifier for the used data link (
                relationship destination).
        :param identifier: Identifier for new output-data link membership.
        :param other_attributes: Optional other attributes as a dictionary or
                list of tuples to be added to the record optionally (default: None).

        """
        return self.new_record(
            PROVONE_OUTPORTTODL,
            identifier,
            {
                PROVONE_ATTR_OUTPUTPORT: out_port,
                PROVONE_ATTR_DATALINK: dt_link,
            },
            other_attributes,
        )

    def DLToOutPort(self, dt_link, out_port, identifier=None, other_attributes=None):
        """

        :param dt_link: Data Link or string identifier for the used data link (
                relationship source).
        :param out_port: Output port or a string identifier for the
                out_port(relationship destination).
        :param identifier: Identifier for new data link-output membership.
        :param other_attributes: Optional other attributes as a dictionary or
                list of tuples to be added to the record optionally (default: None).

        """
        return self.new_record(
            PROVONE_DLTOOUTPORT,
            identifier,
            {
                PROVONE_ATTR_DATALINK: dt_link,
                PROVONE_ATTR_OUTPUTPORT: out_port,
            },
            other_attributes,
        )

    def DLToInPort(self, dt_link, in_port, identifier=None, other_attributes=None):
        """

        :param dt_link: Data Link or string identifier for the used data link (
                relationship source).
        :param in_port: Input port or a string identifier for the in_port (
                relationship destination).
        :param identifier: Identifier for new data link-output membership.
        :param other_attributes: Optional other attributes as a dictionary or
                list of tuples to be added to the record optionally (default: None).

        """
        return self.new_record(
            PROVONE_DLTOINPORT,
            identifier,
            {
                PROVONE_ATTR_DATALINK: dt_link,
                PROVONE_ATTR_INPUTPORT: in_port,
            },
            other_attributes,
        )

    def seqCtrlLink(self, identifier, other_attributes=None):
        """
        Creates a new seq ctrl link.

        :param identifier: Identifier for new seq ctrl link.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).

        """
        return self.new_record(PROVONE_SEQCTRLLINK, identifier, None, other_attributes)

    def control_link_to_process(
        self, used_cntrl_link, used_process, identifier=None, other_attributes=None
    ):
        """

        :param used_cntrl_link: Control Link or a string identifier for the used
        control link (relationship source).
        :param used_process: Data Link or string identifier for the used process
        (relationship destination).
        :param identifier: Identifier for new control link to process relation.
        :param other_attributes: Optional other attributes as a dictionary or
                list of tuples to be added to the record optionally (default: None).

        """
        return self.new_record(
            PROVONE_CLTODESTP,
            identifier,
            {
                PROVONE_ATTR_SEQCTRLLINK: used_cntrl_link,
                PROVONE_ATTR_PROCESS: used_process,
            },
            other_attributes,
        )

    def process_to_control_link(
        self, used_process, used_cntrl_link, identifier=None, other_attributes=None
    ):
        """

        :param used_process: Process or string identifier for the used process
        (relationship source).
        :param used_cntrl_link: Control Link or a string identifier for the used
        control link (relationship destination).
        :param identifier: Identifier for new process to control link relation.
        :param other_attributes: Optional other attributes as a dictionary or
                list of tuples to be added to the record optionally (default: None).

        """
        return self.new_record(
            PROVONE_SOURCEPTOCL,
            identifier,
            {
                PROVONE_ATTR_PROCESS: used_process,
                PROVONE_ATTR_SEQCTRLLINK: used_cntrl_link,
            },
            other_attributes,
        )

    def linkage(
        self,
        data_item,
        dl_link,
        related_process=None,
        identifier=None,
        other_attributes=None,
    ):
        """

        :param data_item: Data or string identifier for the associated data (
        relationship source).
        :param dl_link: Data link or string identifier for the data link (
        relationship destination).
        :param related_process: Process or string identifier of the data link (default=None)
        :param identifier: Identifier for new data-on-link relation.
        :param other_attributes: Optional other attributes as a dictionary or
                list of tuples to be added to the record optionally (default: None).

        """
        return self.new_record(
            PROVONE_DATAONLINK,
            identifier,
            {
                PROVONE_ATTR_DATA: data_item,
                PROVONE_ATTR_DATALINK: dl_link,
                PROVONE_ATTR_PROCESS: related_process,
            },
            other_attributes,
        )

    def parameterization(
        self, in_port, data_item, identifier=None, other_attributes=None
    ):
        """

        :param in_port: InputPort or string identifier for the associated in port (
        relationship source).
        :param data_item: Data item or string identifier for the associated
        default parameter (relationship destination).
        :param identifier: Identifier for new default parameter relation.
        :param other_attributes: Optional other attributes as a dictionary or
                list of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(
            PROVONE_HASDEFAULTPARAM,
            identifier,
            {
                PROVONE_ATTR_INPUTPORT: in_port,
                PROVONE_ATTR_DATA: data_item,
            },
            other_attributes,
        )

    def serialize(self, destination=None, format="json", **args):  # noqa: A002
        """
        Serialize the :py:class:`ProvDocument` to the destination.

        Available serializers can be queried by the value of
        `:py:attr:~prov.serializers.Registry.serializers` after loading them via
        `:py:func:~prov.serializers.Registry.load_serializers()`.

        :param destination: Stream object to serialize the output to. Default is
                `None`, which serializes as a string.
        :param format: Serialization format (default: 'json'), defaulting to
                PROV-JSON.
        :return: Serialization in a string if no destination was given,
                None otherwise.
        """
        serializer = nidm.core.serializers.get(format)(self)
        if destination is None:
            stream = io.StringIO()
            serializer.serialize(stream, **args)
            return stream.getvalue()
        if hasattr(destination, "write"):
            stream = destination
            serializer.serialize(stream, **args)
        else:
            location = destination
            _, netloc, path, _, _, _ = urlparse(location)
            if netloc != "":
                print(
                    "WARNING: not saving as location " + "is not a local file reference"
                )
                return
            fd, name = tempfile.mkstemp()
            stream = os.fdopen(fd, "wb")
            serializer.serialize(stream, **args)
            stream.close()
            if hasattr(shutil, "move"):
                shutil.move(name, path)
            else:
                shutil.copy(name, path)
                os.remove(name)

    def is_part_of(
        self, used_processex, child_processex, identifier=None, other_attributes=None
    ):
        """
        Creates a new is-part-of record for a process exec.

        :param used_processex: The parent processExec (relationship source).
        :param child_processex: The child processExec (relationship destination).
        :param identifier: Identifier for new is-part-of record.
        :param other_attributes: Optional other attributes as a dictionary or list
                of tuples to be added to the record optionally (default: None).
        """
        return self.new_record(
            PROVONE_ISPARTOF,
            identifier,
            {
                PROVONE_ATTR_USED_PREXEC: used_processex,
                PROVONE_ATTR_CHILD_PREXEC: child_processex,
            },
            other_attributes,
        )

    def membership(self, collection, data):
        """
        Creates a new membership record for data to a collection.

        :param collection: Collection the data is to be added to.
        :param data: Data to be added to the collection.
        """
        return self.new_record(
            PROV_MEMBERSHIP,
            None,
            {PROV_ATTR_COLLECTION: collection, PROVONE_ATTR_DATA: data},
        )

    # Aliases
    wasAttributedTo = attribution
    wasAssociatedWith = association
    wasDerivedFrom = derivation
    wasGeneratedBy = generation
    wasInformedBy = communication
    used = usage
    hasInPort = has_in_ports
    hasOutPort = has_out_ports
    hasSubProcess = has_sub_process
    CLtoDestP = control_link_to_process
    sourcePToCL = process_to_control_link
    hasDefaultParam = parameterization
    dataOnLink = linkage
    isPartOf = is_part_of
    hadMember = membership



from ..conftest import BrainVol


def test_project_statistics(brain_vol: BrainVol) -> None:
    AGE_CUTOFF = 30

    project = brain_vol.cmu_test_project_uuid

    # basics stats
    basic_project_stats = brain_vol.restParser.run(
        brain_vol.files, f"/statistics/projects/{project}"
    )
    assert "title" in basic_project_stats

    # basics stats with subjects
    project_stats_with_subjects = brain_vol.restParser.run(
        brain_vol.files, f"/statistics/projects/{project}?fields=subjects"
    )
    assert "title" in project_stats_with_subjects
    assert "subjects" in project_stats_with_subjects

    # filtered subjects stats
    filtered_stats_with_subjects = brain_vol.restParser.run(
        brain_vol.files,
        f"/statistics/projects/{project}?fields=subjects&filter=instruments.AGE_AT_SCAN gt {AGE_CUTOFF}",
    )
    assert "title" in filtered_stats_with_subjects
    assert "subjects" in filtered_stats_with_subjects
    assert len(filtered_stats_with_subjects["subjects"]) < len(
        project_stats_with_subjects["subjects"]
    )

    # filtered subjects instrument stats
    age_stats = brain_vol.restParser.run(
        brain_vol.files,
        f"/statistics/projects/{project}?fields=instruments.AGE_AT_SCAN&filter=instruments.AGE_AT_SCAN gt {AGE_CUTOFF}",
    )
    assert "title" in age_stats
    assert "subjects" in age_stats
    assert len(age_stats["subjects"]) < len(project_stats_with_subjects["subjects"])
    assert "AGE_AT_SCAN" in age_stats
    for x in ["max", "min", "mean", "median", "standard_deviation"]:
        assert x in age_stats["AGE_AT_SCAN"]
    # assert age_stats['AGE_AT_SCAN']['min'] > AGE_CUTOFF
    # assert age_stats['AGE_AT_SCAN']['median'] >= age_stats['AGE_AT_SCAN']['min']
    # assert age_stats['AGE_AT_SCAN']['median'] <= age_stats['AGE_AT_SCAN']['max']

    # filtered subjects instrument and derivative stats
    derivative_stats = brain_vol.restParser.run(
        brain_vol.files,
        f"/statistics/projects/{project}?fields=instruments.AGE_AT_SCAN,derivatives.Right-Hippocampus (mm^3)&filter=instruments.AGE_AT_SCAN gt {AGE_CUTOFF}",
    )
    assert "title" in derivative_stats
    assert "subjects" in derivative_stats
    assert len(derivative_stats["subjects"]) < len(
        project_stats_with_subjects["subjects"]
    )
    for field in ["Right-Hippocampus (mm^3)", "AGE_AT_SCAN"]:
        assert field in derivative_stats
        for x in ["max", "min", "mean", "median", "standard_deviation"]:
            assert x in derivative_stats[field]


def test_project_statistics_fields(brain_vol: BrainVol) -> None:
    project = brain_vol.cmu_test_project_uuid

    # ask for a field based on URI tail
    derivative_stats = brain_vol.restParser.run(
        brain_vol.files,
        f"/statistics/projects/{project}?fields=derivatives.fsl_000020",
    )
    assert "title" in derivative_stats
    assert "subjects" in derivative_stats
    assert len(derivative_stats["subjects"]) > 0
    for field in ["fsl_000020"]:
        assert field in derivative_stats
        for x in ["max", "min", "mean", "median", "standard_deviation"]:
            assert x in derivative_stats[field]

    # ask for a field based on URI tail
    derivative_stats = brain_vol.restParser.run(
        brain_vol.files,
        f"/statistics/projects/{project}?fields=derivatives.fsl_000020,instruments.AGE_AT_SCAN",
    )
    assert "title" in derivative_stats
    assert "subjects" in derivative_stats
    assert len(derivative_stats["subjects"]) > 0
    for field in ["fsl_000020", "AGE_AT_SCAN"]:
        assert field in derivative_stats
        for x in ["max", "min", "mean", "median", "standard_deviation"]:
            assert x in derivative_stats[field]


def test_getTailOfURI(brain_vol: BrainVol) -> None:
    assert (
        brain_vol.restParser.getTailOfURI("http://purl.org/nidash/fsl#fsl_000020")
        == "fsl_000020"
    )
    assert (
        brain_vol.restParser.getTailOfURI("https://surfer.nmr.mgh.harvard.edu/fs_00005")
        == "fs_00005"
    )



from importlib.metadata import version

__version__ = version("pynidm")

try:
    import etelemetry

    etelemetry.check_available_version("incf-nidash/pynidm", __version__)
except ImportError:
    pass



"""
NIDM-Workflows Python API
-------------------------

Python API to create, query, read, and write NIDM-Workflow documents.
"""

from .ProcessExecution import ProcessExecution
from .ProcessSpecification import ProcessSpecification

__all__ = ["ProcessExecution", "ProcessSpecification"]



"""Tools for working with NIDM-Experiment files"""


import re
import click
from rdflib import Graph, Literal, Namespace, URIRef, util
from rdflib.namespace import RDF, split_uri
from nidm.core import Constants
from nidm.experiment.Core import getUUID
from nidm.experiment.Query import GetProjectsMetadata, getProjectAcquisitionObjects
from nidm.experiment.tools.click_base import cli


def is_valid_url_flexible(url):
    """Check if a string is a valid URL, including file:// URLs."""
    url_regex = re.compile(
        r"^(https?|ftp|file):\/\/"  # Allow http, https, ftp, and file schemes
        r"(([a-zA-Z0-9-_.]+)(:[0-9]+)?)?"  # Domain or localhost (optional for file://)
        r"(\/[^\s]*)?$"  # Path (required for file://)
    )
    return bool(url_regex.match(url))


# adding click argument parsing
@cli.command()
@click.option(
    "--nidm_file_list",
    "-nl",
    required=True,
    help="A comma separated list of NIDM files with full path",
)
def update(nidm_file_list):
    """
    This function will update simple 2 NIDM files. It will essentially move some of the metadata in the project activity
    to a prov:Collection entity and then include all nidm:AcquisitionObject entities as prov:hadMember entries in this
    new prov:Collection
    """
    BIDS = Namespace(Constants.BIDS)
    PROV = Namespace(Constants.PROV)
    NIIRI = Namespace(Constants.NIIRI)

    # for each nidm file provided
    for nidm_file in nidm_file_list.split(","):
        # create empty graph
        graph_orig = Graph()

        # bind namespaces
        graph_orig.bind("bids", BIDS)
        graph_orig.bind("prov", PROV)
        graph_orig.bind("niiri", NIIRI)

        # load graph
        graph_orig.parse(nidm_file, format=util.guess_format(nidm_file))

        # create a new prov:Collection, prov:Entity to store metadata about project
        collection_uuid = getUUID()
        graph_orig.add((NIIRI.term(collection_uuid), RDF.type, BIDS.Dataset))
        graph_orig.add((NIIRI.term(collection_uuid), RDF.type, PROV.Collection))
        graph_orig.add((NIIRI.term(collection_uuid), RDF.type, PROV.Entity))

        # query for all metadata in nidm:Project activity
        proj_metadata = GetProjectsMetadata([nidm_file])

        # loop through projects and get associated acquisition objects
        for proj_uuid in proj_metadata["projects"].keys():
            # for all project metadata, only leave dcmitype:title, otherwise move to collection_uuid
            for metadata_key in proj_metadata["projects"][proj_uuid].keys():
                if (metadata_key != str(RDF.type)) and (
                    metadata_key != "dctypes:title"
                ):
                    # get namespace and term from metadata_key
                    nm, term = split_uri(metadata_key)

                    # skip prov namespace as it's already added above
                    if nm.split(":", 1)[0] != "prov":
                        # add RDF namespace
                        temp_nm = Namespace(
                            str(Constants.namespaces[nm.split(":", 1)[0]])
                        )
                        graph_orig.bind(nm.split(":", 1)[0], temp_nm)
                    else:
                        temp_nm = PROV

                    # see if this is a valid URL and encode as one otherwise just a literal
                    if is_valid_url_flexible(
                        proj_metadata["projects"][proj_uuid][metadata_key]
                    ):
                        # now add this metadata to the collection_uuid
                        graph_orig.add(
                            (
                                NIIRI.term(collection_uuid),
                                temp_nm.term(term),
                                URIRef(
                                    proj_metadata["projects"][proj_uuid][metadata_key]
                                ),
                            )
                        )
                    else:
                        # now add this metadata to the collection_uuid
                        graph_orig.add(
                            (
                                NIIRI.term(collection_uuid),
                                temp_nm.term(term),
                                Literal(
                                    proj_metadata["projects"][proj_uuid][metadata_key]
                                ),
                            )
                        )

                    # get namespace and term from proj_uuid
                    proj_nm, proj_term = split_uri(proj_uuid)

                    # graph_orig.remove((proj_uuid, metadata_key,
                    #                   proj_metadata["projects"][proj_uuid][metadata_key]))
                    graph_orig.remove((NIIRI.term(proj_term), temp_nm.term(term), None))
            # get acquisition objects
            acq_objects = getProjectAcquisitionObjects(
                [nidm_file], project_id=str(proj_uuid.split(":", 1)[1])
            )

            for acq_obj in acq_objects:
                graph_orig.add((NIIRI.term(collection_uuid), PROV.hadMember, acq_obj))

        graph_orig.serialize(nidm_file, format="turtle")


if __name__ == "__main__":
    update()



from collections import OrderedDict
from io import StringIO
import json
import random
import re
import string
import uuid
from prov.dot import prov_to_dot
import prov.model as pm
from pydot import Edge
from rdflib import RDF, Graph, URIRef
from ..core import Constants


def getUUID():
    uid = str(uuid.uuid1())
    # added to address some weird bug in rdflib where if the uuid starts with a number, everything up until the first
    # alapha character becomes a prefix...
    if not re.match("^[a-fA-F]+.*", uid):
        # if first digit is not a character than replace it with a randomly selected hex character (a-f).
        uid_temp = uid
        randint = random.randint(0, 5)
        uid = string.ascii_lowercase[randint] + uid_temp[1:]

    return uid


class Core:
    """Base-class for NIDM-Experimenent

    Typically this class is not instantiated directly.  Instantiate one of the child classes such as
    Project, Session, Acquisition, etec.

    @author: David Keator <dbkeator@uci.edu>
    @copyright: University of California, Irvine 2017

    """

    language = "en"

    def __init__(self):
        """
        Default constructor, loads empty graph and namespaces from NIDM/Scripts/Constants.py
        """
        # a new instance of NIDMDocument PROV document with namespaces already bound
        self.graph = Constants.NIDMDocument()
        # make a local copy of the namespaces
        self.namespaces = Constants.namespaces
        # storage for uuid
        self._uuid = None

    # class constructor with user-supplied PROV document/graph, namespaces from Constants.py
    @classmethod
    def withGraph(cls, graph):
        """
        Alternate constructor, loads user-supplied graph and default namespaces from NIDM/Scripts/Constants.py

        Keyword arguments:
            graph -- an rdflib.Graph object
        """
        cls.graph = graph
        cls.namespaces = {}
        # bind namespaces to cls.graph
        for name, namespace in cls.namespaces.items():
            cls.graph.add_namespace(name, namespace)

    # class constructor with user-supplied graph and namespaces
    @classmethod
    def withGraphAndNamespaces(cls, graph, namespaces):
        """
        Alternate constructor, loads user-supplied graph and binds user-supplied namespaces

        :param graph: an rdflib.Graph object
        :param namespaces: python dictionary {namespace_identifier, URL}
        :return: none
        """

        cls.graph = graph
        cls.namespaces = namespaces
        # bind namespaces to cls.graph
        for name, namespace in cls.namespaces.items():
            cls.graph.add_namespace(name, namespace)

    def find_namespace_with_uri(self, uri):
        for namespace in self.graph.namespaces:
            if str(namespace.uri) == str(uri):
                return namespace
        else:
            return False

    def get_uuid(self):
        """
        returns UUID of self
        :return:
        """
        return self._uuid

    def getGraph(self):
        """
        Returns rdflib.Graph object
        """
        return self.graph

    def getNamespace(self):
        """
        Returns namespace dictionary {namespace_id, URL}
        """
        return self.graph._namespaces

    def addNamespace(self, prefix, uri):
        """
        Adds namespace to self.graph
        :param prefix: namespace prefix
        :param uri: namespace URI
        :return: none
        """
        self.graph.add_namespace(prefix, uri)

    def checkNamespacePrefix(self, prefix):
        """
        Checks if namespace prefix already exists in self.graph
        :param prefix: namespace identifier
        :return: True if prefix exists, False if not
        """
        # check if prefix already exists
        return bool(prefix in self.graph._namespaces)

    def safe_string(self, string):
        return (
            string.strip()
            .replace(" ", "_")
            .replace("-", "_")
            .replace(",", "_")
            .replace("(", "_")
            .replace(")", "_")
            .replace("'", "_")
            .replace("/", "_")
            .replace("#", "num")
        )

    def getDataType(self, var):
        if isinstance(var, int):
            return pm.XSD_INTEGER
        elif isinstance(var, float):
            return pm.XSD_FLOAT
        elif isinstance(var, str):
            return pm.XSD_STRING
        elif isinstance(var, list):
            return list
        else:
            print("datatype not found...")
            return None

    def add_person(self, uuid=None, attributes=None, add_default_type=True):
        """
        Simply adds prov:agent to graph and returns object
        :param role:
        :param attributes:
        :return:
        """

        if uuid is not None:
            # add Person agent with existing uuid
            person = self.graph.agent("niiri:" + uuid, other_attributes=attributes)
        else:
            # add Person agent
            person = self.graph.agent("niiri:" + getUUID(), other_attributes=attributes)

        if add_default_type:
            # add minimal attributes to person
            person.add_attributes({pm.PROV_TYPE: pm.PROV["Person"]})

        # connect self to person serving as role
        # if(isinstance(self,pm.ProvActivity)):
        #    self.wasAssociatedWith(person)
        # elif(isinstance(self,pm.ProvEntity)):
        #    self.wasAttributedTo(person)

        return person

    def add_qualified_association(
        self, person, role, plan=None, attributes=None  # noqa: U100
    ):
        """
        Adds a qualified association to self object
        :param person: prov:agent to associated
        :param role: prov:hadRole to associate
        :param plan: optional prov:hadPlan to associate
        :param attributes: optional attributes to add to qualified association
        :return: association
        """

        # connect self to person serving as role
        # WIP this doesn't work for subclasses as they don't have the pm.ProvActivity type
        # Might be able to use the following and look into the tuples but for now skip this check
        # import inspect
        # class_tree = inspect.getclasstree([self.__class__])

        # if(isinstance(self, pm.ProvActivity)):

        # associate person with activity for qualified association
        assoc = self.graph.association(
            agent=person,
            activity="niiri:" + self.get_uuid(),
            other_attributes={pm.PROV_ROLE: role},
        )

        # add wasAssociatedWith association
        # self.wasAssociatedWith(person)

        return assoc

    def addLiteralAttribute(
        self, namespace_prefix, term, object, namespace_uri=None  # noqa: A002
    ):
        """
        Adds generic literal and inserts into the graph
        :param namespace_prefix: namespace prefix
        :param pred_term: predidate term to associate with tuple
        :param object: literal to add as object of tuple
        :param namespace_uri: If namespace_prefix isn't one already used then use this optional argument to define
        :return: none
        """
        # figure out datatype of literal
        datatype = self.getDataType(object)
        # check if namespace prefix already exists in graph
        if not self.checkNamespacePrefix(namespace_prefix):
            # if so, use URI
            # namespace_uri = self.namespaces[namespace_prefix]
            # else: #add namespace_uri + prefix to graph
            if namespace_uri is None:
                raise TypeError(
                    "Namespace_uri argument must be defined for new namespaces"
                )
            else:
                self.addNamespace(namespace_prefix, namespace_uri)

        # figure out if predicate namespace is defined, if not, return predicate namespace error
        try:
            if datatype is not None:
                self.add_attributes(
                    {
                        str(namespace_prefix + ":" + term): pm.Literal(
                            object, datatype=datatype
                        )
                    }
                )
            else:
                self.add_attributes(
                    {str(namespace_prefix + ":" + term): pm.Literal(object)}
                )
        except KeyError as e:
            print(
                '\nPredicate namespace identifier "',
                str(e).split("'")[1],
                '" not found! \n',
            )
            print(
                "Use addNamespace method to add namespace before adding literal attribute \n"
            )
            print("No attribute has been added \n")

    def addAttributesWithNamespaces(self, id, attributes):  # noqa: A002
        """
        Adds generic attributes in bulk to object [id] and inserts into the graph

        :param id: subject identifier/URI
        :param attributes: List of dictionaries with keys prefix, uri, term, value} \
        example: [ {uri:"http://ncitt.ncit.nih.gov/", prefix:"ncit", term:"age", value:15},
                   {uri:"http://ncitt.ncit.nih.gov/", prefix:"ncit", term:"gender", value:"M"}]
        :return: TypeError if namespace prefix already exists in graph but URI is different
        """
        # iterate through list of attributes
        for tple in attributes:
            # check if namespace prefix already exists in graph
            if self.checkNamespacePrefix(tple["prefix"]):
                # checking if existing prefix maps to same namespaceURI, if so use it, if not then raise error
                if self.namespaces[tple["prefix"]] != tple["uri"]:
                    raise TypeError(
                        "Namespace prefix: "
                        + tple["prefix"]
                        + "already exists in document"
                    )

            else:  # add tuple to graph
                self.addNamespace(tple["prefix"], tple["uri"])

            # figure out datatype of literal
            datatype = self.getDataType(tple["value"])
            if datatype is not None:
                id.add_attributes(
                    {
                        self.graph.namespaces[tple["prefix"]][tple["term"]]: pm.Literal(
                            tple["value"], datatype=datatype
                        )
                    }
                )
            else:
                id.add_attributes(
                    {
                        self.graph.namespaces[tple["prefix"]][tple["term"]]: pm.Literal(
                            tple["value"]
                        )
                    }
                )

    def addAttributes(self, id, attributes):  # noqa: A002
        """
        Adds generic attributes in bulk to object [id] and inserts into the graph

        :param id: subject identifier/URI
        :param attributes: Dictionary with keys as prefix:term and value of attribute} \
        example: {"ncit:age":15,"ncit:gender":"M", Constants.NIDM_FAMILY_NAME:"Keator"}
        :return: TypeError if namespace prefix does not exist in graph
        """
        # iterate through attributes
        for key in attributes.keys():
            # is the key already mapped to a URL (i.e. using one of the constants from Constants.py) or is it in prefix:term form?
            # if not validators.url(key):
            # check if namespace prefix already exists in graph or #if we're using a Constants reference
            # these should be of type pm.QualifiedName otherwise they might be a string version of a qualified name
            # which should be changed but for now we're supporting these types that look like "nidm:Derivative"
            if isinstance(key, pm.QualifiedName):
                # check if the namespace is in self.graph
                ns = self.find_namespace_with_uri(str(key.namespace.uri))
                if ns is False:
                    # namespace isn't in graph so give an error
                    raise TypeError(
                        "Namespace prefix "
                        + key.prefix
                        + ", "
                        + key.namespace.uri
                        + " not in graph, use addAttributesWithNamespaces or manually add!"
                    )
            elif isinstance(key, str):
                if not self.checkNamespacePrefix(key.split(":")[0]):
                    raise TypeError(
                        "Namespace prefix "
                        + key
                        + " not in graph, use addAttributesWithNamespaces or manually add!"
                    )
            # figure out datatype of literal
            datatype = self.getDataType(attributes[key])
            # if (not validators.url(key)):
            # we must be using the prefix:term form instead of a constant directly

            #    if (datatype != None):
            #        id.add_attributes({self.namespaces[key.split(':')[0]][key.split(':')[1]]:Literal(attributes[key],datatype=datatype)})
            #    else:
            #        id.add_attributes({self.namespaces[key.split(':')[0]][key.split(':')[1]]:Literal(attributes[key])})
            # else:
            # we're using the Constants form
            if datatype is not None:
                id.add_attributes({key: pm.Literal(attributes[key], datatype=datatype)})
            else:
                id.add_attributes({key: pm.Literal(attributes[key])})

    def get_metadata_dict(self, NIDM_TYPE):
        """
        This function converts metadata to a dictionary using uris as keys
        :param NIDM_TYPE: a prov qualified name type (e.g. Constants.NIDM_PROJECT, Constants.NIDM_SESSION, etc.)
        :return: dictionary object containing metadata

        """
        # create empty project_metadata json object
        metadata = {}

        # use RDFLib here for temporary graph making query easier
        rdf_graph = Graph()

        rdf_graph_parse = rdf_graph.parse(
            source=StringIO(self.serializeTurtle()), format="turtle"
        )

        # get subject uri for object

        uri = None
        for s in rdf_graph_parse.subjects(
            predicate=RDF.type, object=URIRef(NIDM_TYPE.uri)
        ):
            uri = s

        if uri is None:
            print(f"Error finding {NIDM_TYPE} in NIDM-Exp Graph")
            return metadata

        # Cycle through metadata and add to json
        for predicate, objects in rdf_graph.predicate_objects(subject=uri):
            metadata[str(predicate)] = str(objects)

        return metadata

    def serializeTurtle(self):
        """
        Serializes graph to Turtle format
        :return: text of serialized graph in Turtle format
        """
        return self.graph.serialize(None, format="rdf", rdf_format="ttl")

    def serializeTrig(self, identifier=None):
        """
        Serializes graph to Turtle format
        :param identifier: Optional identifier to use for graph serialization
        :return: text of serialized graph in Turtle format
        """
        if identifier is not None:
            rdf_graph = Graph(identifier=identifier)
            rdf_graph.parse(source=StringIO(self.serializeTurtle()), format="turtle")
        else:
            rdf_graph = Graph()
            rdf_graph.parse(source=StringIO(self.serializeTurtle()), format="turtle")

        # return rdf_graph.serialize(format='trig').decode('ASCII')
        return rdf_graph.serialize(format="trig")

    def serializeJSONLD(self):
        """
        Serializes graph to JSON-LD format
        :return: text of serialized graph in JSON-LD format
        """
        # workaround to get JSONLD from RDFLib...
        rdf_graph = Graph()
        # rdf_graph_parse = rdf_graph.parse(source=StringIO(self.serializeTurtle()),format='turtle')
        rdf_graph_parse = rdf_graph.parse(
            source=StringIO(self.graph.serialize(None, format="rdf", rdf_format="ttl")),
            format="turtle",
        )

        # WIP: currently this creates a default JSON-LD context from Constants.py and not in the correct way from the
        # NIDM-E OWL files that that will be the next iteration
        context1 = self.createDefaultJSONLDcontext()
        # This part adds to the context any prefixes in an existing NIDM-E file that might have been added by a user
        # and isn't covered by the default namespaces / constants in Constants.py
        context2 = self.prefix_to_context()
        context = dict(context1, **context2)

        # WIP: LOOK AT https://github.com/satra/nidm-jsonld
        # return rdf_graph_parse.serialize(format='json-ld', context=context, indent=4).decode('ASCII')
        # g=rdf_graph_parse.serialize(format='json-ld', indent=4).decode('ASCII')
        g = rdf_graph_parse.serialize(format="json-ld", indent=4)

        import pyld as ld

        return json.dumps(ld.jsonld.compact(json.loads(g), context), indent=4)

    def createDefaultJSONLDcontext(self):
        """
        This function returns a context dictionary for NIDM-E JSON serializations
        :return: context dictionary
        """

        from nidm.core.Constants import namespaces
        from nidm.experiment.Utils import load_nidm_owl_files

        # load current OWL files
        load_nidm_owl_files()

        context = {}

        context["@version"] = 1.1
        context["records"] = {}
        context["records"]["@container"] = "@type"
        context["records"]["@id"] = "@graph"

        # load Constants.namespaces
        context.update(Constants.namespaces)

        context.update(
            {
                "xsd": {"@type": "@id", "@id": "http://www.w3.org/2001/XMLSchema#"},
                "prov": {"@type": "@id", "@id": "http://www.w3.org/ns/prov#"},
                "agent": {"@type": "@id", "@id": "prov:agent"},
                "entity": {"@type": "@id", "@id": "prov:entity"},
                "activity": {"@type": "@id", "@id": "prov:activity"},
                "hadPlan": {"@type": "@id", "@id": "prov:hadPlan"},
                "hadRole": {"@type": "@id", "@id": "prov:hadRole"},
                "wasAttributedTo": {"@type": "@id", "@id": "prov:wasAttributedTo"},
                "association": {"@type": "@id", "@id": "prov:qualifiedAssociation"},
                "usage": {"@type": "@id", "@id": "prov:qualifiedUsage"},
                "generation": {"@type": "@id", "@id": "prov:qualifiedGeneration"},
                "startedAtTime": {"@type": "xsd:dateTime", "@id": "prov:startedAtTime"},
                "endedAtTime": {"@type": "xsd:dateTime", "@id": "prov:endedAtTime"},
            }
        )

        # add namespaces from Constants.namespaces
        for key, value in namespaces.items():
            # context['@context'][key] = value
            context[key] = value

        # add terms from Constants.nidm_experiment_terms
        for term in Constants.nidm_experiment_terms:
            # context['@context'][term.localpart] = term.uri
            context[term.localpart] = term.uri

        # add prefix's from current document...this accounts for new terms
        context.update(self.prefix_to_context())

        # WIP
        # cycle through OWL graph and add terms
        # For anything that has a label

        # for s, o in sorted(term_graph.subject_objects(Constants.RDFS['label'])):
        #    json_key = str(o)
        #    if '_' in json_key:
        #        json_key = str(o).split('_')[1]
        #    context['@context'][json_key] = OrderedDict()

        #    if s in term_graph.ranges:
        #        context['@context'][json_key]['@id'] = str(s)
        #        context['@context'][json_key]['@type'] = next(iter(term_graph.ranges[s]))
        #    else:
        #        context['@context'][json_key] = str(s)

        return context

    def save_DotGraph(self, filename, format=None):  # noqa: A002
        dot = prov_to_dot(self.graph)

        ISPARTOF = {
            "label": "isPartOf",
            "fontsize": "10.0",
            "color": "darkgreen",
            "fontcolor": "darkgreen",
        }
        style = ISPARTOF

        # query self.graph for Project uuids
        # use RDFLib here for temporary graph making query easier
        rdf_graph = Graph()
        rdf_graph = rdf_graph.parse(
            source=StringIO(self.graph.serialize(None, format="rdf", rdf_format="ttl")),
            format="turtle",
        )

        # SPARQL query to get project UUIDs
        query = """
        PREFIX nidm:<http://purl.org/nidash/nidm#>
        PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>

        SELECT distinct ?uuid
        Where {
            {
                ?uuid rdf:type nidm:Project
            }

        }
        """
        qres = rdf_graph.query(query)
        for row in qres:
            print(f"project uuid = {row}")
            # parse uuid from project URI
            # project_uuid = str(row[0]).rsplit('/', 1)[-1]
            project_uuid = str(row[0])
            # for each Project uuid search dot structure for Project uuid
            project_node = None
            for key, value in dot.obj_dict["nodes"].items():
                # get node number in DOT graph for Project
                if project_uuid in str(value[0]["attributes"].get("URL", "")):
                    project_node = key
                    break

        # for each Session in Project class self.sessions list, find node numbers in DOT graph

        for session in self.sessions:
            print(session)
            for key, value in dot.obj_dict["nodes"].items():
                # get node number in DOT graph for Project
                if session.identifier.uri in str(value[0]["attributes"].get("URL", "")):
                    session_node = key
                    # print(f"session node = {key}")

                    # add to DOT structure edge between project_node and session_node
                    dot.add_edge(Edge(session_node, project_node, **style))

                    # for each Acquisition in Session class ._acquisitions list, find node numbers in DOT graph
                    for acquisition in session.get_acquisitions():
                        # search through the nodes again to figure out node number for acquisition
                        for key, value in dot.obj_dict["nodes"].items():
                            # get node number in DOT graph for Project
                            if acquisition.identifier.uri in str(
                                value[0]["attributes"].get("URL", "")
                            ):
                                acquisition_node = key
                                # print(f"acquisition node = {key}")

                                dot.add_edge(
                                    Edge(acquisition_node, session_node, **style)
                                )
        # for each derivative activity, add edge to project for isPartOf
        for derivative_activity in self._derivatives:
            print(derivative_activity)
            for key, value in dot.obj_dict["nodes"].items():
                # get node number in DOT graph for Project
                if derivative_activity.identifier.uri in str(
                    value[0]["attributes"].get("URL", "")
                ):
                    derivative_activity_node = key
                    # print(f"session node = {key}")

                    # add to DOT structure edge between project_node and session_node
                    dot.add_edge(Edge(derivative_activity_node, project_node, **style))

        # add some logic to find nodes with dct:hasPart relation and add those edges to graph...prov_to_dot ignores these
        if format != "None":
            dot.write(filename, format=format)
        else:
            dot.write(filename, format="pdf")

    def prefix_to_context(self):
        """
        This function returns a context dictionary for JSONLD export from current NIDM-Exp document....
        :return: Context dictionary for JSONLD
        """

        # This sets up basic contexts from namespaces in documents
        context = OrderedDict()
        for key, value in self.graph._namespaces.items():
            # context[key] = {}
            # context[key]['@type']='@id'
            # context[key]['@id']= value.uri

            # context[key]['@type']='@id'

            # added for some weird namespaces where key is URIRef and value is Namespace
            # seems to only apply to PROV and NIDM qualified names.
            # has something to do with read_nidm function in Utils and add_metadata_for_subject
            # when it comes across a NIDM or PROV term.
            if isinstance(key, URIRef):
                continue
            else:
                context[key] = str(value.uri)

        # This adds suffix part of namespaces as IDs to make things read easier in JSONLD
        # for namespace in self.graph.namespaces:
        #    context[namespace.qname()]='@id'
        #    context[namespace.qname()]=namespace.qname().localpart

        return context

    def __str__(self):
        return "NIDM-Experiment Base Class"



from nidm.experiment.tools import (  # noqa: F401
    nidm_concat,
    nidm_convert,
    nidm_linreg,
    nidm_merge,
    nidm_query,
    nidm_version,
    nidm_visualize,
)
from nidm.experiment.tools.click_base import cli

__all__ = ["cli"]



from __future__ import annotations
from pathlib import Path
import requests


def urlretrieve(url: str, filepath: str | Path) -> None:
    """`requests`-based alternative to `urllib.request.urlretrieve`"""
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(filepath, "wb") as fp:
            for chunk in r.iter_content(65535):
                fp.write(chunk)



"""
NIDM-Experiment Python API
--------------------------

Python API to create, query, read, and write NIDM-Experiment documents.
"""

from .Acquisition import Acquisition
from .AcquisitionObject import AcquisitionObject
from .AssessmentAcquisition import AssessmentAcquisition
from .AssessmentObject import AssessmentObject
from .Core import Core
from .DataElement import DataElement
from .DemographicsObject import DemographicsObject
from .Derivative import Derivative
from .DerivativeObject import DerivativeObject
from .MRAcquisition import MRAcquisition
from .MRObject import MRObject
from .PETAcquisition import PETAcquisition
from .PETObject import PETObject
from .Project import Project
from .Session import Session

__all__ = [
    "Acquisition",
    "AcquisitionObject",
    "AssessmentAcquisition",
    "AssessmentObject",
    "Core",
    "DataElement",
    "DemographicsObject",
    "Derivative",
    "DerivativeObject",
    "MRAcquisition",
    "MRObject",
    "PETAcquisition",
    "PETObject",
    "Project",
    "Session",
]
